"""An intake driver for FDB/GSV access"""
import os
import fnmatch
import datetime
import requests
import eccodes
import xarray as xr
import numpy as np
import dask
from ruamel.yaml import YAML
from intake.source import base
from aqua.core.util.eccodes import get_eccodes_attr
from aqua.core.util import to_list
from aqua.core.logger import log_configure, _check_loglevel
from .timeutil import check_dates, shift_time_dataset, floor_datetime, read_bridge_date, todatetime
from .timeutil import split_date, make_timeaxis, date2str, date2yyyymm, add_offset

# Test if FDB5 binary library is available
try:
    from gsv.retriever import GSVRetriever
    gsv_available = True
except RuntimeError:
    gsv_available = False
    gsv_error_cause = "FDB5 binary library not present on system or outdated"
except KeyError:
    gsv_available = False
    gsv_error_cause = "Environment variables for gsv, such as GRID_DEFINITION_PATH, not set."

BRIDGE_API_URL = "https://qubed.lumi.apps.dte.destination-earth.eu/api/v2/stac"  # LUMI QUBED STAC API


class GSVSource(base.DataSource):
    container = 'xarray'
    name = 'gsv'
    version = '0.0.2'
    partition_access = True

    _ds = None  # _ds and _da will contain samples of the data for dask access
    _da = None
    dask_access = False  # Flag if dask has been requested
    first_run = True  # Flag to check if this is the first run of the class

    def __init__(self, request, data_start_date, data_end_date, bridge_start_date=None, bridge_end_date=None, 
                 hpc_expver=None, timestyle="date",
                 chunks="S", savefreq="h", timestep="h", timeshift=None,
                 startdate=None, enddate=None, var=None, metadata=None, level=None,
                 switch_eccodes=False, loglevel='WARNING', engine='fdb', **kwargs):
        """
        Initializes the GSVSource class. These are typically specified in the catalog entry,
        but can also be specified upon accessing the catalog.

        Args:
            request (dict): Request dictionary
            data_start_date (str): Start date of the available data.
            data_end_date (str): End date of the available data.
            bridge_end_date (str, optional): End date of the bridge data. Defaults to None.
            bridge_start_date (str, optional): Start date of the bridge data. Defaults to None.
            hpc_expver (str, optional): Alternative expver to be used if the data are on hpc
            timestyle (str, optional): Time style. Defaults to "date".
            chunks (str or dict, optional): Time and vertical chunking.
                                        If a string is provided, it is assumed to be time chunking.
                                        If it is a dictionary the keys 'time' and 'vertical' are looked for.
                                        Time chunking can be one of S (step), 10M, 15M, 30M, h, 1h, 3h, 6h, D, 5D, W, M, Y.
                                        Defaults to "S".
                                        Vertical chunking is expressed as the number of vertical levels to be used.
                                        Defaults to None (no vertical chunking).
            timestep (str, optional): Time step. Can be one of 10M, 15M, 30M, 1h, h, 3h, 6h, D, 5D, W, M, Y.
                                      Defaults to "h".
            startdate (str, optional): Start date for request. Defaults to None.
            enddate (str, optional): End date for request. Defaults to None.
            var (str, optional): Variable ID. Defaults to those in the catalog.
            metadata (dict, optional): Metadata read from catalog. Contains path to FDB.
            level (int, float, list): optional level(s) to be read. Must use the same units as the original source.
            switch_eccodes (bool, optional): Flag to activate switching of eccodes path. Defaults to False.
            engine (str, optional): Engine to be used for GSV retrieval: 'polytope' or 'fdb'. Defaults to 'fdb'. 
            loglevel (string) : The loglevel for the GSVSource
            kwargs: other keyword arguments.
        """

        self.logger = log_configure(log_level=loglevel, log_name='GSVSource')
        self.engine = engine
        self.gsv_log_level = _check_loglevel(self.logger.getEffectiveLevel())
        self.logger.debug("Init of the GSV source class")

        if not gsv_available:
            raise ImportError(gsv_error_cause)

        self._request = request.copy()

        if metadata:
            self.fdbhome = metadata.get('fdb_home', None)
            self.fdbpath = metadata.get('fdb_path', None)
            self.fdbhome_bridge = metadata.get('fdb_home_bridge', None)
            self.fdbpath_bridge = metadata.get('fdb_path_bridge', None)
            if switch_eccodes:
                self.eccodes_path = metadata.get('eccodes_path', None)
                self.logger.info("ECCODES switching to %s", self.eccodes_path)
            else:
                self.logger.debug("ECCODES switching is off")
                self.eccodes_path = None
            self.levels = metadata.get('levels', None)
            self.fdb_info_file = metadata.get('fdb_info_file', None)

            # safety check for paths

            # skip the first time:
            # this is needed because intake calls initialization twice, first without custom arguments and then with
            # If we pass engine='polytope' on a remote machine the path check would fail but intake initializes the class a first time actually with the wrong engine
            if self.engine and self.engine == 'fdb' and not GSVSource.first_run:
                for attr in ['fdbhome', 'fdbpath', 'fdbhome_bridge', 
                            'fdbpath_bridge', 'eccodes_path']:
                    attr_path = getattr(self, attr)
                    if attr_path and not os.path.exists(attr_path):
                        raise FileNotFoundError(f'{attr} path {attr_path} does not exist!')

        else:
            self.fdbpath = None
            self.fdbhome = None
            self.fdbhome_bridge = None
            self.fdbpath_bridge = None
            self.fdb_info_file = None
            self.eccodes_path = None
            self.levels = None

        GSVSource.first_run = False

        # set the timestyle
        self.timestyle = timestyle
        self.timeshift = timeshift
        self.itime = 0  # position of time dim

        self.ilevel = None

        if not var:  # if no var provided keep the default in the catalog
            self._var = request["param"]
        else:
            self._var = var

        self._var = to_list(self._var)  # Make sure self._var is a list

        # Convert var names to paramId. The usage of strings is discouraged, so a warning is issued
        for i, v in enumerate(self._var):
            if isinstance(v, str):
                self.logger.warning("Variable %s is a string, conversion to paramid may lead to errors", v)
                self._var[i] = int(get_eccodes_attr(v)['paramId'])

        self.logger.debug("List of paramid to retrieve %s", self._var)

        self._kwargs = kwargs
        self.hpc_expver = hpc_expver

        # set all the start/end dates for data and bridge
        self.data_start_date = None
        self.data_end_date = None
        self.bridge_start_date = None
        self.bridge_end_date = None

        self._define_start_end_dates(data_start_date, data_end_date, bridge_start_date, bridge_end_date)
        # set all the start/end dates for the retrieval
        self._define_retrieve_dates(startdate, enddate)

        # flooring to the frequency the time to ensure that hourly, daily and monthly data
        # are read at the right time frequency
        # setting hpc and bridge availability dates
        for attr in ["data_start_date", "data_end_date", "bridge_end_date", 
                     "bridge_start_date", "startdate", "enddate"]:
            setattr(self, attr, floor_datetime(getattr(self, attr), savefreq))

        self.logger.debug('Data frequency (i.e. savefreq): %s', savefreq)
        self.logger.debug('Data_start_date: %s, Data_end_date: %s, Bridge_start_date: %s, Bridge_end_date: %s',
                          self.data_start_date, self.data_end_date, self.bridge_start_date, self.bridge_end_date)
        self.logger.debug('Request startdate: %s, Request enddate: %s', self.startdate, self.enddate)

        if self.timestyle != "yearmonth":
            offset = int(self._request.get("step", 0))  # optional initial offset for steps (in timesteps)

            # special for 6h: set offset startdate if needed
            self.startdate = add_offset(data_start_date, self.startdate, offset, timestep)

        if isinstance(chunks, dict):
            chunking_time = chunks.get('time', 'S')
            chunking_vertical = chunks.get('vertical', None)
        else:
            chunking_time = chunks
            chunking_vertical = None

        if chunking_time.upper() == "S":  # special case: time chunking is single saved frame
            chunking_time = savefreq

        self.data_startdate, self.data_starttime = split_date(self.data_start_date)

        if "levelist" in self._request:
            levelist = to_list(self._request["levelist"])
            if level:
                level = to_list(level)
                idx = list(map(levelist.index, level))
                self.idx_3d = idx
                self._request["levelist"] = level  # override default levels
                if self.levels:  # if levels in metadata select them too
                    self.levels = to_list(self.levels)
                    self.levels = [self.levels[i] for i in idx]
            else:
                self.idx_3d = list(range(0, len(levelist)))
        else:
            self.idx_3d = None

        self.onelevel = False
        if "levelist" in self._request:
            if self.levels:  # Do we have physical levels specified in metadata?
                lev = self._request["levelist"]
                if isinstance(lev, list) and len(lev) > 1:
                    self.onelevel = True  # If yes we can afford to read only one level
            else:
                self.logger.warning("A speedup of data retrieval could be achieved by specifying the levels keyword in metadata.")

        timeaxis = make_timeaxis(self.data_start_date, self.startdate, self.enddate,
                                 shiftmonth=self.timeshift, timestep=timestep,
                                 savefreq=savefreq, chunkfreq=chunking_time,
                                 bridge_start_date=self.bridge_start_date, bridge_end_date=self.bridge_end_date)

        self.timeaxis = timeaxis["timeaxis"]
        self.chk_start_idx = timeaxis["start_idx"]
        self.chk_start_date = timeaxis["start_date"]
        self.chk_end_idx = timeaxis["end_idx"]
        self.chk_end_date = timeaxis["end_date"]
        self.chk_size = timeaxis["size"]
        self._npartitions = len(self.chk_start_date)
        self.chk_type = timeaxis["type"]

        if not np.array_equal(self.chk_type, timeaxis["type_end"]):  # sanity check
            raise ValueError('Chunk size is not aligned with bridge_start_data and bridge_end_data. Fix your catalog!')
        if np.any(self.chk_type == 0):
            if not self.fdbpath and not self.fdbhome:
                raise ValueError('Some data is on HPC but no local FDB path or FDB home is specified in catalog.')
        if np.any(self.chk_type == 1):
            if not self.fdbpath_bridge and not self.fdbhome_bridge:
                raise ValueError('Some data is on bridge but no bridge FDB path or FDB home specified in catalog.')

        self.chk_vert = None
        self.ntimechunks = self._npartitions
        self.nlevelchunks = None

        if "levelist" in self._request:
            self.chunking_vertical = chunking_vertical
            if self.chunking_vertical:
                levelist = to_list(self._request["levelist"])
                if len(levelist) <= self.chunking_vertical:
                    self.chunking_vertical = None
                else:
                    self.chk_vert = [levelist[i:i+self.chunking_vertical] for i in range(0, len(levelist),
                                                                                         self.chunking_vertical)]
                    self.ntimechunks = self._npartitions
                    self.nlevelchunks = len(self.chk_vert)
                    self._npartitions = self._npartitions*len(self.chk_vert)
        else:
            self.chunking_vertical = None  # no vertical chunking

        self._switch_eccodes()

        super(GSVSource, self).__init__(metadata=metadata)

    def _define_start_end_dates(self, data_start_date, data_end_date, bridge_start_date, bridge_end_date):
        """
        Define the start and end dates of the data and bridge

        Args:
            data_start_date (str): Start date of the available data.
            data_end_date (str): End date of the available data.
            bridge_start_date (str): Start date of the bridge data.
            bridge_end_date (str): End date of the bridge data.
        """
        # Getting info from the FDB info file
        fdb_info = self._read_fdb_info()
        
        # Data dates
        self._setup_data_dates(data_start_date, data_end_date, fdb_info)
        
        # Bridge dates
        self._setup_bridge_dates(bridge_start_date, bridge_end_date, fdb_info)
        self._adjust_bridge_bounds()

    def _read_fdb_info(self):
        """
        Read FDB information from file if available
        
        Returns:
            dict or None: FDB information if available, None otherwise
        """
        if self.fdb_info_file:
            self.logger.debug('Reading FDB info from file %s', self.fdb_info_file)
            return self.get_fdb_definitions_from_file(self.fdb_info_file)
        return None

    def _setup_data_dates(self, data_start_date, data_end_date, fdb_info):
        """
        Setup data start and end dates
        
        Args:
            data_start_date (str): Start date of the available data.
            data_end_date (str): End date of the available data.
            fdb_info (dict or None): FDB information if available
        """
        if self.fdb_info_file and fdb_info:
            self.data_start_date = fdb_info['data']['data_start_date']
            self.data_end_date = fdb_info['data']['data_end_date']
            self.hpc_expver = fdb_info['data']['expver']
        else:
            if data_start_date == 'auto' or data_end_date == 'auto':
                self.logger.debug('Autoguessing of the FDB start and end date enabled.')
                if self.timestyle == 'yearmonth':
                    raise ValueError('Auto date selection not supported for timestyle=yearmonth. Please specify start and end date!')
                self.data_start_date, self.data_end_date = self.parse_fdb(data_start_date, data_end_date)
            else:
                self.data_start_date = data_start_date
                self.data_end_date = data_end_date

    def _setup_bridge_dates(self, bridge_start_date, bridge_end_date, fdb_info):
        """
        Setup bridge start and end dates
        
        Args:
            bridge_start_date (str): Start date of the bridge data.
            bridge_end_date (str): End date of the bridge data.
            fdb_info (dict or None): FDB information if available
        """
        if self.fdb_info_file and fdb_info and fdb_info.get('bridge'):
            self._setup_bridge_dates_from_file(fdb_info)
        else:
            if bridge_start_date == 'stac' or bridge_end_date == 'stac':
                self._setup_bridge_dates_from_stac()
            else:
                self._setup_bridge_dates_from_input(bridge_start_date, bridge_end_date)

    def _setup_bridge_dates_from_file(self, fdb_info):
        """
        Setup bridge dates from FDB info file
        
        Args:
            fdb_info (dict): FDB information from file
        """
        self.bridge_start_date = fdb_info['bridge']['bridge_start_date']
        self.bridge_end_date = fdb_info['bridge']['bridge_end_date']
        self._request['expver'] = fdb_info['bridge']['expver']

    def _setup_bridge_dates_from_stac(self):
        """
        Setup bridge dates from STAC API
        """
        self.logger.debug('Reading FDB info from bridge STAC API')
        self.bridge_start_date, self.bridge_end_date = self.get_dates_from_stac_api(self._request, BRIDGE_API_URL)
        self.bridge_end_date = self.bridge_end_date + 'T2300'
        self.bridge_start_date = self.bridge_start_date + 'T0000'
        self.logger.debug('STAC API bridge start data: %s, bridge end date: %s', 
                        self.bridge_start_date, self.bridge_end_date)

    def _setup_bridge_dates_from_input(self, bridge_start_date, bridge_end_date):
        """
        Setup bridge dates from input parameters
        
        Args:
            bridge_start_date (str): Start date of the bridge data.
            bridge_end_date (str): End date of the bridge data.
        """
        # deprecated method that guess from text file and fall back
        self.bridge_start_date = read_bridge_date(bridge_start_date)
        self.bridge_end_date = read_bridge_date(bridge_end_date)

    def _adjust_bridge_bounds(self):
        """
        Adjust bridge bounds if not specified or set to 'complete'
        """
        if self.bridge_start_date == 'complete' or self.bridge_end_date == 'complete':
            self.bridge_start_date = self.data_start_date
            self.bridge_end_date = self.data_end_date
        if not self.bridge_start_date and self.bridge_end_date:
            self.bridge_start_date = self.data_start_date
        if not self.bridge_end_date and self.bridge_start_date:
            self.bridge_end_date = self.data_end_date

    def _define_retrieve_dates(self, startdate, enddate):
        """
        Define the start and end dates for the retrieval
        """

        if not startdate:
            self.startdate = self.data_start_date
        else:
            self.startdate = startdate

        if not enddate:
            self.enddate = self.data_end_date
        else:
            self.enddate = enddate

    def __getstate__(self):
        """
        This helps to pickle variables which were added later, after initialization of the class.
        These are all self variables needed by get_partition().
        """

        return {
            'data_startdate': self.data_startdate,
            'data_starttime': self.data_starttime,
            'chk_start_idx': self.chk_start_idx,
            'chk_end_idx': self.chk_end_idx,
            'chk_start_date': self.chk_start_date,
            'chk_end_date': self.chk_end_date,
            'chunking_vertical': self.chunking_vertical,
            'chk_vert': self.chk_vert,
            'chk_type': self.chk_type,
            'hpc_expver': self.hpc_expver,
            '_request': self._request,
            'timestyle': self.timestyle,
            'fdbhome': self.fdbhome,
            'fdbpath': self.fdbpath,
            'fdbhome_bridge': self.fdbhome_bridge,
            'fdbpath_bridge': self.fdbpath_bridge,
            'eccodes_path': self.eccodes_path,
            '_var': self._var,
            'timeshift': self.timeshift,
            'gsv_log_level': self.gsv_log_level,
            'logger': self.logger,
            'engine': self.engine
        }

    def __setstate__(self, state):
        """
        This helps to restore from pickle variables which were added later after initialization.
        These are all self variables needed by get_partition().
        """

        self.data_startdate = state['data_startdate']
        self.data_starttime = state['data_starttime']
        self.chk_start_idx = state['chk_start_idx']
        self.chk_end_idx = state['chk_end_idx']
        self.chk_start_date = state['chk_start_date']
        self.chk_end_date = state['chk_end_date']
        self.chunking_vertical = state['chunking_vertical']
        self.chk_vert = state['chk_vert']
        self.chk_type = state['chk_type']
        self.hpc_expver = state['hpc_expver']
        self.timestyle = state['timestyle']
        self.fdbhome = state['fdbhome']
        self.fdbpath = state['fdbpath']
        self.fdbhome_bridge = state['fdbhome_bridge']
        self.fdbpath_bridge = state['fdbpath_bridge']
        self.eccodes_path = state['eccodes_path']
        self._var = state['_var']
        self.timeshift = state['timeshift']
        self._request = state['_request']
        self.gsv_log_level = state['gsv_log_level']
        self.logger = state['logger']
        self.engine = state['engine']

    def _get_schema(self):
        """
        Standard method providing data schema.
        For dask access it is assumed that all DataArrays read share the same shape and data type.
        """

        # check if dates are within acceptable range
        check_dates(self.startdate, self.data_start_date, self.enddate, self.data_end_date)

        if self.dask_access:  # We need a better schema for dask access
            if not self._ds or not self._da:  # we still have to retrieve a sample dataset

                self._ds = self._get_partition(0, var=self._var, first=True, onelevel=self.onelevel)

                var = list(self._ds.data_vars)[0]
                da = self._ds[var]  # get first variable dataarray

                # If we have multiple levels, then this array needs to be expanded
                if self.onelevel:
                    lev = self.levels
                    apos = da.dims.index("level")  # expand the size of the "level" axis
                    attrs = da["level"].attrs
                    da = da.squeeze("level").drop_vars("level").expand_dims(level=lev, axis=apos)
                    da["level"].attrs.update(attrs)

                self._da = da

            metadata = {
                'dims': self._da.dims,
                'attrs': self._ds.attrs
            }
            schema = base.Schema(
                datashape=None,
                dtype=str(self._da.dtype),
                shape=da.shape,
                name=None,
                npartitions=self._npartitions,
                extra_metadata=metadata)
        else:
            schema = base.Schema(
                datashape=None,
                dtype=str(xr.Dataset),
                shape=None,
                name=None,
                npartitions=self._npartitions,
                extra_metadata={},
            )

        return schema

    def _switch_eccodes(self):
        """
        Internal method to switch ECCODES version if needed.
        """
        if self.eccodes_path:  # if needed switch eccodes path
            # unless we have already switched
            if self.eccodes_path and (self.eccodes_path != eccodes.codes_definition_path()):
                eccodes.codes_context_delete()  # flush old definitions in cache
                eccodes.codes_set_definitions_path(self.eccodes_path)

    def _index_to_timelevel(self, ii):
        """
        Internal method to convert partition index to time and level indices
        """
        if self.chunking_vertical:
            i = ii // len(self.chk_vert)
            j = ii % len(self.chk_vert)
        else:
            i = ii
            j = 0
        return i, j

    def _get_partition(self, ii, var=None, first=False, onelevel=False):
        """
        Standard internal method reading i-th data partition from FDB

        Args:
            ii (int): partition number
            var (string, optional): single variable to retrieve. Defaults to using those set at init
            first (bool, optional): read only the first step (used for schema retrieval)
            onelevel (bool, optional): read only one level. Defaults to False.

        Returns:
            An xarray.DataSet
        """

        request = self._request.copy()  # We are going to change it, threads do need this

        i, j = self._index_to_timelevel(ii)
        if self.chunking_vertical:
            request["levelist"] = self.chk_vert[j]

        if self.timestyle == "date":
            dds, tts = date2str(self.chk_start_date[i])
            dde, tte = date2str(self.chk_end_date[i])
            if ((dds == dde) and (tts == tte)) or first:
                request["date"] = f"{dds}"
                request["time"] = f"{tts}"
            else:
                request["date"] = f"{dds}/to/{dde}"
                request["time"] = f"{tts}/to/{tte}"

        elif self.timestyle == "step":  # style is 'step'
            request["date"] = self.data_startdate
            request["time"] = self.data_starttime
            s0 = self.chk_start_idx[i]
            s1 = self.chk_end_idx[i]
            if s0 == s1 or first:
                request["step"] = f'{s0}'
            else:
                request["step"] = f'{s0}/to/{s1}'

        elif self.timestyle == "yearmonth":  # style is 'yearmonth'
            yys, mms = date2yyyymm(self.chk_start_date[i])
            yye, mme = date2yyyymm(self.chk_end_date[i])
            if ((yys == yye) or first):
                request["year"] = f"{yys}"
            else:
                request["year"] = f"{yys}/to/{yye}"
            if ((mms == mme) or first):
                request["month"] = f"{mms}"
            else:
                request["month"] = f"{mms}/to/{mme}"
            # HACK: step is required by the code, but not needed by GSV
            # for key in ["date", "step", "time"]:
            #    if key in request:
            #        del request[key]
        else:
            raise ValueError(f'Timestyle {self.timestyle} not supported')

        if onelevel:  # limit to one single level
            request["levelist"] = request["levelist"][0]

        # If a var is used and it is a string, it means that previous parts of the code have failed
        # to convert it to paramId. The conversion is then asked to GSVRetriever, which rely on a
        # dictionary of paramId to shortName. This is not guaranteed to work, so a warning is issued.
        if var:
            if isinstance(var, str):
                self.logger.warning("Asking for variable %s as string, this may lead to errors", var)
            request["param"] = var
        else:
            request["param"] = self._var

        # Select based on type of FDB
        fstream_iterator = False  # We set it False, but it works also with True

        if self.chk_type[i]:
            # Bridge FDB type
            if self.fdbhome_bridge:
                os.environ["FDB_HOME"] = self.fdbhome_bridge
                self.logger.debug('Access is BRIDGE and FDB_HOME is set to %s', self.fdbhome_bridge)
            if self.fdbpath_bridge:
                os.environ["FDB5_CONFIG_FILE"] = self.fdbpath_bridge
                self.logger.debug('Access is BRIDGE and FDB5_CONFIG_FILE is set to %s', self.fdbpath_bridge)
            fstream_iterator = True
        else:
            # HPC FDB type
            if self.fdbhome:  # if fdbhome is provided, use it, since we are creating a new gsv
                os.environ["FDB_HOME"] = self.fdbhome
                self.logger.debug('Access is HPC and FDB_HOME is set to %s', self.fdbhome)
            if self.fdbpath:  # if fdbpath provided, use it, since we are creating a new gsv
                os.environ["FDB5_CONFIG_FILE"] = self.fdbpath
                self.logger.debug('Access is HPC and FDB5_CONFIG_FILE is set to %s', self.fdbpath)
            if self.hpc_expver:
                request["expver"] = self.hpc_expver

        self._switch_eccodes()

        # The following is a hack around a pyfdb/fdb5 bug which requires a double initialization when reading from bridge
        # See https://github.com/DestinE-Climate-DT/AQUA/issues/1715
        # Notice also that for some mysterious reason this works only if the result is stored in self (even if then it is not used)
        if self.chk_type[i]:
            self.gsv = GSVRetriever(engine=self.engine, logging_level=self.gsv_log_level)
        gsv = GSVRetriever(engine=self.engine, logging_level=self.gsv_log_level)

        self.logger.debug('Request %s', request)
        dataset = gsv.request_data(request, use_stream_iterator=fstream_iterator,
                                   process_derived_variables=False)  # following 2.9.2 we avoid derived variables

        if self.timeshift:  # shift time by one month (special case)
            dataset = shift_time_dataset(dataset)

        return dataset

    def read(self):
        """Return a in-memory dask dataset"""
        ds = [self._get_partition(i) for i in range(self._npartitions)]
        ds = xr.concat(ds, dim='time', coords='different')
        return ds

    def get_part_delayed(self, ii, var, shape, dtype):
        """
        Function to read a delayed partition.
        Returns a dask.array

        Args:
            ii (int): partition number
            var (string): variable name
            shape: shape of the schema
            dtype: data type of the schema
        """

        i, j = self._index_to_timelevel(ii)

        ds = dask.delayed(self._get_partition)(ii, var=var)

        # get the data from the first (and only) data array
        ds = ds.to_array()[0].data
        newshape = list(shape)
        newshape[self.itime] = self.chk_size[i]
        if self.chunking_vertical:  # if we have vertical chunking
            newshape[self.ilevel] = len(self.chk_vert[j])

        return dask.array.from_delayed(ds, newshape, dtype)

    def to_dask(self):
        """Return a dask xarray dataset for this data source"""

        self.dask_access = True  # This is used to tell _get_schema() to load dask info
        self._load_metadata()

        shape = self._schema.shape
        dtype = self._schema.dtype

        self.itime = self._da.dims.index("time")
        if self.chunking_vertical:
            self.ilevel = self._da.dims.index("level")

        if 'valid_time' in self._da.coords:  # temporary hack because valid_time is inconsistent anyway
            self._da = self._da.drop_vars('valid_time')

        coords = self._da.coords.copy()
        coords['time'] = self.timeaxis

        ds = xr.Dataset()

        # Now works only with the variables which have been read (the fixer may change names later)
        # Notice that the mismatch between shortnames in different versions of eccodes is handled here
        # We consider stable between versions the paramId, not the shortName. This means that we read
        # the GRIB_paramid attribute and based on this we get the shortName from the eccodes definitions.
        # If you want to read the shortName according to previous versions of eccodes, you need to
        # set the switch_eccodes flag to True in the catalog.
        for var in self._ds.data_vars:
            # We need to ask for the GRIB_paramid that is attribute of the variable
            original_paramid = self._ds[var].attrs.get("GRIB_paramId", var)
            updated_var = get_eccodes_attr(original_paramid)['shortName']
            # If this is executed, it means that the shortname came from a previous version of eccodes
            # and we're using a more recent one in which the paramId is still existing, but the shortName
            # has changed. This is a warning to the user. However the final variable name will be influenced
            # by this only if fix=False.
            if updated_var != var:
                self.logger.warning("Variable shortname %s has been interpreted with another eccodes. Current eccodes %s will read paramid %s as %s", var, eccodes.__version__, original_paramid, updated_var)
            # Create a dask array from a list of delayed get_partition calls
            if not self.chunking_vertical:
                dalist = [self.get_part_delayed(i, original_paramid, shape, dtype) for i in range(self.npartitions)]
                darr = dask.array.concatenate(dalist, axis=self.itime)  # This is a lazy dask array
            else:
                dalist = []
                for j in range(self.nlevelchunks):
                    dalistlev = [self.get_part_delayed(i*self.nlevelchunks+j, original_paramid, shape, dtype) for i in range(self.ntimechunks)]
                    dalist.append(dask.array.concatenate(dalistlev, axis=self.itime))
                darr = dask.array.concatenate(dalist, axis=self.ilevel)  # This is a lazy dask array

            da = xr.DataArray(darr,
                              name=updated_var,
                              attrs=self._ds[var].attrs, # We need the original var to retrieve the attributes
                              dims=self._da.dims,
                              coords=coords)

            log_history(da, "Dataset retrieved by GSV interface")

            ds[updated_var] = da

        ds.attrs.update(self._ds.attrs)
        if self.idx_3d:
            ds = ds.assign_coords(idx_level=("level", self.idx_3d))

        return ds

    # Overload read_chunked() from base.DataSource
    def read_chunked(self):
        """Return iterator over container fragments of data source"""
        self._load_metadata()
        for i in range(self.npartitions):
            ds = self._get_partition(i)
            if self.idx_3d:
                ds = ds.assign_coords(idx_level=("level", self.idx_3d))
            yield ds

    def get_fdb_definitions_from_file(self, fdb_info_file):
        """
        Get the FDB definitions from a file

        Args:
            file (str): path to the file

        Returns:
            dict: definitions
        """
        if not os.path.exists(fdb_info_file):
            self.logger.error("FDB info file %s does not exist", fdb_info_file)
            return None

        yaml = YAML()

        try:
            with open(fdb_info_file, 'r') as file:
                fdb_info = yaml.load(file)
        except (OSError, yaml.YAMLError) as e:
            self.logger.error("Error reading or parsing YAML file %s: %s", fdb_info_file, str(e))
            return None

        # The 'data' block is mandatory and present since the first chunck of simulation
        # The 'bridge' block is written only if some data is on bridge (see #1760)
        if 'data' in fdb_info:
            try:
                fdb_info['data']['data_start_date'] = self._validate_info_date(fdb_info, 'data', 'start')
                fdb_info['data']['data_end_date'] = self._validate_info_date(fdb_info, 'data', 'end')
            except KeyError:
                self.logger.error("FDB info file %s does not contain HPC dates in correct format", fdb_info_file)
                return None
        else:
            self.logger.error("FDB info file %s does not contain 'data' section, which is mandatory", fdb_info_file)
            return None
        if 'bridge' in fdb_info and (self.fdbhome_bridge or self.fdbpath_bridge):
            try:
                fdb_info['bridge']['bridge_start_date'] = self._validate_info_date(fdb_info, 'bridge', 'start')
                fdb_info['bridge']['bridge_end_date'] = self._validate_info_date(fdb_info, 'bridge', 'end')
            # if bridge dates are wrongly defined, set the bridge block to None
            except KeyError:
                self.logger.error("FDB info file %s does not contain bridge dates in correct form", fdb_info_file)
                fdb_info['bridge'] = None
        else:
            fdb_info['bridge'] = None

        return fdb_info

    @staticmethod
    def _validate_info_date(fdb_info_file, location='data', kind='start'):

        if location not in ['data', 'bridge']:
            raise ValueError(f'location {location} should be either data or local')

        if kind not in ['start', 'end']:
            raise ValueError(f'kind {kind} should be either start or end')

        return todatetime(fdb_info_file[location][f'{location}_{kind}_date']).strftime('%Y%m%dT%H%M')

    def parse_fdb(self, start_date, end_date):
        """
        Parse the FDB config file and return the start and end dates of the data.
        This works only with the DE GSV schema.

        Args:
            start_date (str): if 'auto' the start date is found automatically. Else it is the start date.
            end_date (str): if 'auto' the end date is found automatically. Else it is the end date.

        Returns:
            tuple: start and end dates
        """
        if not self.fdbhome and not self.fdbpath:
            raise ValueError('Automatic dates requested but no FDB home or FDB path specified in catalog.')

        yaml = YAML()

        if self.fdbhome:  # FDB_HOME takes precedence but assumes a fixed subdirectory structure
            yamlfile = os.path.join(self.fdbhome, 'etc/fdb/config.yaml')
        else:
            yamlfile = self.fdbpath

        with open(yamlfile, 'r') as file:
            cfg = yaml.load(file)

        if 'fdbs' in cfg:
            root = cfg['fdbs'][0]['spaces'][0]['roots'][0]['path']
        else:
            root = cfg['spaces'][0]['roots'][0]['path']

        req = self._request
        expver = req['expver']
        if self.hpc_expver:
            expver = self.hpc_expver

        file_mask = f"{req['class']}:{req['dataset']}:{req['activity']}:{req['experiment']}:{req['generation']}:{req['model']}:{req['realization']}:{expver}:{req['stream']}:*"

        file_mask = file_mask.lower()
        file_list = [
            f for f in os.listdir(root) if fnmatch.fnmatch(f.lower(), file_mask)
        ]

        datesel = [filename[-8:] for filename in file_list if (filename[-8:].isdigit() and len(filename[-8:]) == 8)]
        datesel.sort()

        if len(datesel) == 0:
            raise ValueError('Auto date selection in catalog but no valid dates found in FDB')
        else:
            if start_date == 'auto':
                start_date = datesel[0] + 'T0000'
            if end_date == 'auto':
                end_date = datesel[-2] + 'T2300'
            self.logger.info('Automatic FDB date range: %s - %s', start_date, end_date)

        return start_date, end_date
    
    @staticmethod
    def get_dates_from_stac_api(params, base_url=BRIDGE_API_URL):
        """
        Function to get from the STAC data bridge the available
        dates of a dataset on the bridge
        
        Args:
            params (dict): Dictionary of parameters to interrogate the STAC API.
                        In principle, the same as the usual FDB request
            base_url (str): URL for the STAC API

        Returns:
            tuple: A tuple containing the start and end dates of the dataset
        """

        # Define the base URL for the STAC API
        params['root'] = 'root'
        params['param'] = to_list(params['param'])[0]
        for p in ['date', 'time', 'step', 'year', 'month']:
            params.pop(p, None)  # remove date/time/step/year/month if present

        # new stac API requires lowercased keys
        params = {k: v.lower() if isinstance(v, str) else v for k, v in params.items()}

        # network problems can happen, so we need to handle them
        try:
            response = requests.get(base_url, params=params, timeout=10)
        except requests.Timeout as e:
            raise TimeoutError("STAC API request timed out after 10 seconds.") from e
        except requests.RequestException as e:
            raise ConnectionError("STAC API request failed") from e

        # Check the response status code
        if response.status_code == 400:
            raise ValueError(f"Bad request to STAC API: {response.text}")
        if response.status_code == 503:
            raise ValueError(f"Service unavailable: {response.text}")
        if response.status_code != 200:
            raise ValueError(f"Unexpected response from STAC API: {response.status_code} - {response.text}")

        # parse the JSON response
        try:
            stac_json = response.json()
        except ValueError as exc:
            raise ValueError("Failed to parse STAC API response as JSON") from exc

    
        dateblock = [el for el in stac_json['links'] if el.get('title') == 'date']
        if not dateblock:
            raise ValueError(f"The first link in the response is not a date link, but {dateblock}")

        # specific extraction of the dates: new format following the qube STAC API
        dates = dateblock[0].get('variables').get('date').get('enum')
        sorted_dates = sorted(dates)
        
        return sorted_dates[0], sorted_dates[-1]


# This function is repeated here in order not to create a cross dependency between GSVSource and AQUA
def log_history(data, msg):
    """Elementary provenance logger in the history attribute"""

    if isinstance(data, (xr.DataArray, xr.Dataset)):
        now = datetime.datetime.now()
        date_now = now.strftime("%Y-%m-%d %H:%M:%S")
        hist = data.attrs.get("history", "") + f"{date_now} {msg};\n"
        data.attrs.update({"history": hist})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
AQUA catalog operations mixin
'''

import os
import sys
import shutil
from urllib.error import HTTPError

import fsspec

from aqua.core.lock import SafeFileLock
from aqua.core.util import load_yaml, dump_yaml
from aqua.core.util.util import to_list, HiddenPrints
from aqua.core.reader.catalog import show_catalog_content as print_catalog

# folder used for reading/storing catalogs
CATPATH = 'catalogs'


class CatalogMixin:
    """Mixin for AQUA catalog operations"""

    def set(self, args):
        """Set an installed catalog as the one used in the config-aqua.yaml

        Args:
            args (argparse.Namespace): arguments from the command line
        """

        self._check()

        if os.path.exists(f"{self.configpath}/{CATPATH}/{args.catalog}"):
            self._set_catalog(args.catalog)
        else:
            self.logger.error('%s catalog is not installed!', args.catalog)
            sys.exit(1)


    def add(self, args):
        """Add a catalog and set it as a default in config-aqua.yaml

        Args:
            args (argparse.Namespace): arguments from the command line
        """
        print('Adding the AQUA catalog', args.catalog)
        self._check(silent=True)

        if args.editable is not None:
            self._add_catalog_editable(args.catalog, args.editable)
        else:
            self._add_catalog_github(args.catalog, args.repository)

        # verify that the new catalog is compatible with AQUA, loading it with catalog()
        try:
            with HiddenPrints():
                print_catalog()
        except Exception as e:
            self.remove(args)
            self.logger.error('Current catalog is not compatible with AQUA, removing it for safety!')
            self.logger.error(e)
            sys.exit(1)

    def _add_catalog_editable(self, catalog, editable):
        """Add a catalog in editable mode (i.e. link)"""

        cdir = f'{self.configpath}/{CATPATH}/{catalog}'
        editable = os.path.abspath(editable)
        print('Installing catalog in editable mode from', editable, 'to', self.configpath)
        if os.path.exists(editable):
            if os.path.exists(cdir):
                self.logger.error('Catalog %s already installed in %s, please consider `aqua remove`',
                                  catalog, cdir)
                sys.exit(1)
            else:
                os.symlink(editable, cdir)
        else:
            self.logger.error('Catalog %s cannot be found in %s', catalog, editable)
            sys.exit(1)

        self._set_catalog(catalog)

    def _github_explore(self, repository=None):
        """
        Explore the remote GitHub repository

        Args:
            repository (str): the repository to explore, if None it uses the default
                              DestinE-Climate-DT/Climate-DT-catalog

        Returns:
            fsspec.filesystem: the filesystem object for the GitHub repository
        """
        if repository is None:
            org = 'DestinE-Climate-DT'
            repo = 'Climate-DT-catalog'
        else:
            try:
                org, repo = repository.split('/')
            except ValueError:
                self.logger.error('Repository should be in the format user/repo, got %s', repository)
                sys.exit(1)
        try:
            # Detect if running in GitHub Actions
            is_github_actions = os.getenv("GITHUB_ACTIONS") == "true"
            token = os.getenv("GITHUB_TOKEN")  # Get GitHub token

            # Use authentication only when running inside GitHub Actions
            username = 'github-actions' if is_github_actions else os.getenv("GITHUB_USER")
            if token and username:
                auth_kwargs = {"username": username, "token": token}
                self.logger.info("Using authenticated access to GitHub API.")
            else:
                auth_kwargs = {}
                self.logger.warning("Running without authentication. Rate limits may apply.")
                self.logger.warning("Consider setting GITHUB_TOKEN and GITHUB_USER environment variables for authenticated access.")

            fs = fsspec.filesystem(
                "github",
                org=org,
                repo=repo,
                **auth_kwargs  # Apply authentication if available
            )
            self.logger.info('Accessed remote repository https://github.com/%s/%s', org, repo)
        except HTTPError:
            self.logger.error('Permission issues in accessing Climate-DT catalog, please contact AQUA maintainers')
            sys.exit(1)

        return fs

    def avail(self, args):
        """
        Return the catalog available on the Github website

        Args:
            args (argparse.Namespace): arguments from the command line
        """
        fs = self._github_explore(repository=args.repository)
        available_catalog = [os.path.basename(x) for x in fs.ls(f"{CATPATH}/")]
        print('Available ClimateDT catalogs at are:')
        print(available_catalog)

    def _add_catalog_github(self, catalog, repository=None):
        """
        Add a catalog from a remote Github repository.
        Default repository is the Climate-DT repository
        DestinE-Climate-DT/Climate-DT-catalog

        Args:
            catalog (str): the catalog to be added
            repository (str): the repository from which to fetch the catalog, if None it uses the default
        """
        # recursive copy
        cdir = f'{self.configpath}/{CATPATH}/{catalog}'
        if not os.path.exists(cdir):
            fs = self._github_explore(repository)
            available_catalog = [os.path.basename(x) for x in fs.ls(f"{CATPATH}/")]
            if catalog not in available_catalog:
                self.logger.error('Cannot find on %s the requested catalog %s, available are %s',
                                  repository if repository else 'DestinE-Climate-DT/Climate-DT-catalog',
                                  catalog, available_catalog)
                sys.exit(1)

            source_dir = f"{CATPATH}/{catalog}"
            self.logger.info('Fetching remote catalog %s from github to %s', catalog, cdir)
            os.makedirs(cdir, exist_ok=True)
            self._fsspec_get_recursive(fs, source_dir, cdir)
            self.logger.info('Download complete!')
            self._set_catalog(catalog)
        else:
            self.logger.error("Catalog %s already installed in %s, please consider `aqua update`.",
                              catalog, cdir)
            sys.exit(1)

    def _update_catalog(self, catalog):
        """Update a catalog by copying it if not installed in editable mode

        Args:
            catalog (str): the catalog to be updated
        """
        cdir = f'{self.configpath}/{CATPATH}/{catalog}'
        if os.path.exists(cdir):
            if os.path.islink(cdir):
                self.logger.error('%s catalog has been installed in editable mode, no need to update', catalog)
                sys.exit(1)
            self.logger.info('Removing %s from %s', catalog, cdir)
            shutil.rmtree(cdir)
            self._add_catalog_github(catalog)
        else:
            self.logger.error('%s does not appear to be installed, please consider `aqua add`', catalog)
            sys.exit(1)

    def _set_catalog(self, catalog):
        """Modify the config-aqua.yaml with the proper catalog

        Args:
            catalog (str): the catalog to be set as the default in the config-aqua.yaml
        """

        self.logger.info('Setting catalog name to %s', catalog)
        with SafeFileLock(self.configfile + '.lock', loglevel=self.loglevel):
            cfg = load_yaml(self.configfile)
            if cfg['catalog'] is None:
                self.logger.debug('No catalog previously installed: setting catalog name to %s', catalog)
                cfg['catalog'] = catalog
            else:
                if catalog not in to_list(cfg['catalog']):
                    self.logger.debug('Adding catalog %s to the existing list %s', catalog, cfg['catalog'])
                    cfg['catalog'] = [catalog] + to_list(cfg['catalog'])
                else:
                    if isinstance(cfg['catalog'], list):
                        other_catalogs = [x for x in cfg['catalog'] if x != catalog]
                        self.logger.debug('Catalog %s is already there, setting it as first entry before %s',
                                        catalog, other_catalogs)
                        cfg['catalog'] = [catalog] + other_catalogs
                    else:
                        self.logger.debug('Catalog %s is already there, but is the only installed', catalog)
                        cfg['catalog'] = catalog

            dump_yaml(self.configfile, cfg)

    def remove(self, args):
        """Remove a catalog

        Args:
            args (argparse.Namespace): arguments from the command line
        """
        self._check()
        if '/' in args.catalog:
            args.catalog = os.path.basename(args.catalog)
        cdir = f'{self.configpath}/{CATPATH}/{args.catalog}'
        print('Remove the AQUA catalog', args.catalog, 'from', cdir)
        if os.path.exists(cdir):
            if os.path.islink(cdir):
                os.unlink(cdir)
            else:
                shutil.rmtree(cdir)
            self._clean_catalog(args.catalog)
        else:
            self.logger.error('Catalog %s is not installed in %s, cannot remove it',
                              args.catalog, cdir)
            sys.exit(1)

    def _clean_catalog(self, catalog):
        """
        Remove catalog from the configuration file
        """

        with SafeFileLock(self.configfile + '.lock', loglevel=self.loglevel):
            cfg = load_yaml(self.configfile)
            if isinstance(cfg['catalog'], str):
                cfg['catalog'] = None
            else:
                cfg['catalog'].remove(catalog)
            self.logger.info('Catalog %s removed, catalogs %s are available', catalog, cfg['catalog'])
            dump_yaml(self.configfile, cfg)

    @staticmethod
    def _fsspec_get_recursive(fs, src_dir, dest_dir):
        """
        Recursive function to download from a fsspec object

        Args:
            fs: fsspec filesystem object, as github instance
            src_dir (str): source directory
            dest_dir (str): target directory

        Returns:
            Remotely copy data from source to dest directory
        """
        data = fs.ls(src_dir)
        for item in data:
            relative_path = os.path.relpath(item, src_dir)
            dest_path = os.path.join(dest_dir, relative_path)

            if fs.isdir(item):
                # Create the directory in the destination
                os.makedirs(dest_path, exist_ok=True)
                # Recursively copy the contents of the directory
                CatalogMixin._fsspec_get_recursive(fs, item, dest_path)
            else:
                # Ensure the directory exists before copying the file
                os.makedirs(os.path.dirname(dest_path), exist_ok=True)
                # Copy the file
                fs.get(item, dest_path)



"""Tests for the GridBuilder class."""
import subprocess
import os
import pytest
from aqua import GridBuilder
from aqua import Reader
from aqua.core.gridbuilder.gridentrymanager import GridEntryManager
from aqua.core.configurer import ConfigPath
from aqua.core.util import load_yaml

pytestmark = [pytest.mark.aqua, pytest.mark.xdist_group(name="grid_builder")]


class TestGridBuilder:
    """Test the GridBuilder class."""
    grid_dir = f'{ConfigPath().configdir}/grids'

    def test_grid_healpix_polytope_atm2d(self, tmp_path):
        """Test the GridBuilder class with a HEALPix grid."""
        reader = Reader(
            model="IFS-FESOM", exp="story-2017-control", source="hourly-hpz7-atm2d",
            engine="polytope", areas=False, chunks={'time': 'H'})
        data = reader.retrieve(var='2t')
        grid_builder = GridBuilder(outdir=tmp_path, model_name='IFS', original_resolution='tco1279')
        grid_builder.build(data, verify=True, create_yaml=False)

    def test_grid_healpix_polytope_ocean2d(self, tmp_path):
        """Test the GridBuilder class with a HEALPix grid and rebuild option."""
        reader = Reader(
            model="IFS-FESOM", exp="story-2017-control", source="daily-hpz7-oce2d",
            engine="polytope", areas=False, chunks={'time': 'D'})
        data = reader.retrieve(var='tos')
        grid_builder = GridBuilder(outdir=tmp_path, model_name='FESOM', original_resolution='ng5')
        grid_builder.build(data, verify=True, create_yaml=False)

    @pytest.mark.parametrize("rebuild", [False, True])
    def test_grid_regular(self, tmp_path, rebuild):
        """Test the GridBuilder class with a regular grid."""
        reader = Reader(model='IFS', exp='test-tco79', source='long', loglevel='debug', areas=False, fix=False)
        data = reader.retrieve()
        grid_builder = GridBuilder(outdir=tmp_path, original_resolution='tco79')
        grid_builder.build(data, verify=True, rebuild=rebuild, create_yaml=True)
        assert os.path.exists(f'{self.grid_dir}/regular.yaml')
    
    def test_grid_curvilinear(self, tmp_path):
        """Test the GridBuilder class with a regular grid."""
        reader = Reader(model='ECE4-FAST', exp='test', source='monthly-oce', loglevel='debug', areas=False)
        data = reader.retrieve()
        grid_builder = GridBuilder(outdir=tmp_path, model_name='nemo', grid_name='ORCA2')
        grid_builder.build(data, verify=False, create_yaml=True) #set to False since it is very heavy
        assert os.path.exists(f'{self.grid_dir}/nemo-curvilinear.yaml')
        grid = load_yaml(f'{self.grid_dir}/nemo-curvilinear.yaml')
        assert set(grid['grids']['nemo-ORCA2-2d']['space_coord']) == set(['y', 'x'])
        assert grid['grids']['nemo-ORCA2-2d']['remap_method'] == 'bil'

    def test_grid_unstructured(self, tmp_path):
        """Test the GridBuilder class with an unstructured grid."""
        reader = Reader(model='ECE4-FAST', exp='test', source='monthly-atm', loglevel='debug', areas=False)
        data = reader.retrieve()
        grid_builder = GridBuilder(outdir=tmp_path, model_name='ifs', grid_name='tl63')
        grid_builder.build(data, verify=True, create_yaml=True) # this is not working yet
        assert os.path.exists(f'{self.grid_dir}/ifs-unstructured.yaml')
    
    def test_grid_healpix(self, tmp_path):
        """Test the GridBuilder class with a HEALPix grid."""
        reader = Reader(model='ERA5', exp='era5-hpz3', source='monthly', loglevel='debug', areas=False)
        data = reader.retrieve()
        grid_builder = GridBuilder(outdir=tmp_path, original_resolution='N320')
        grid_builder.build(data, verify=True, create_yaml=False)


class TestGridEntryManager:
    """Test the GridEntryManager class."""
    @pytest.mark.parametrize(
        "model,aquagrid,cdogrid,original,kind", [
            ('IFS', 'hpz7-nested', 'hp32_nested', 'tco79', 'healpix'),
            ('IFS', 'tco79', None, None, 'unstructured'),
            ('IFS', 'r100', 'r180x90', 'tco79', 'regular'),
            ('NEMO','eORCA1', None, None, 'curvilinear'),
        ]
    )
    def test_gem_basic(self, model, aquagrid, cdogrid, original, kind):
        """Test the GridEntryManager class without mask"""
        gem = GridEntryManager(
            model_name=model, original_resolution=original, grid_name=aquagrid
        )
        filename = gem.get_gridfilename(cdogrid, kind)
        basename = gem.get_basename(aquagrid, cdogrid)
        if cdogrid:
            assert f'{kind}.yaml' == os.path.basename(filename)
            assert aquagrid == basename
        else:
            assert f'{model.lower()}-{kind}.yaml' == os.path.basename(filename)
            assert f'{model.lower()}_{aquagrid}' == basename
       
        entry = gem.create_grid_entry_name(aquagrid, cdogrid)
        assert basename.replace('_', '-') == entry

    def test_gem_curvilinear(self):
        """Test the GridEntryManager class with a curvilinear grid."""
        gem = GridEntryManager(model_name='nemo', grid_name='ORCA2')

        filename = gem.get_gridfilename(cdogrid=None, gridkind='curvilinear')
        assert os.path.basename(filename) == 'nemo-curvilinear.yaml'
        basename = gem.get_basename(aquagrid='orca2', cdogrid=None, masked='oce', vert_coord='depth')
        assert basename == 'nemo_orca2_3d_depth'
        entry = gem.create_grid_entry_name(aquagrid='orca2', cdogrid=None, masked='oce', vert_coord='depth')
        assert 'nemo-orca2-3d-depth' == entry
    
        block = gem.create_grid_entry_block(
            path='orca2_oce_depth_v1.nc',
            horizontal_dims='cells',
            cdo_options='-f nc',
            remap_method='bil',
            vert_coord='depth'
        )
        assert block['space_coord'] == 'cells'
        assert block['vert_coord'] == 'depth'
        assert block['cdo_options'] == '-f nc'
        assert block['remap_method'] == 'bil'
        assert block['path']['depth'] == 'orca2_oce_depth_v1.nc'




"""YAML utility functions"""

import os
from string import Template as DefaultTemplate
from collections import defaultdict
from tempfile import TemporaryDirectory
from jinja2 import Template
from ruamel.yaml import YAML
import yaml  # This is needed to allow YAML override in intake
from aqua.core.logger import log_configure

def construct_yaml_merge(loader, node):
    """
    This function is used to enable override in yaml for intake
    """
    if isinstance(node, yaml.ScalarNode):
        # Handle scalar nodes
        return loader.construct_scalar(node)
    
    # Handle sequence nodes
    maps = []
    for subnode in node.value:
        maps.append(loader.construct_object(subnode))
    result = {}
    for dictionary in reversed(maps):
        result.update(dictionary)
    return result


# Run this to enable YAML override for the yaml package when using SafeLoader in intake 
yaml.SafeLoader.add_constructor(
            'tag:yaml.org,2002:merge',
            construct_yaml_merge)


def load_multi_yaml(folder_path: str | None = None, filenames: list | None = None,
                    definitions: str | dict | None = None, **kwargs):
    """
    Load and merge yaml files.
    If a filenames list of strings is provided, only the yaml files with
    the matching full path will be merged.
    If a folder_path is provided, all the yaml files in the folder will be merged.

    Args:
        folder_path (str, optional): the path of the folder containing the yaml
                                        files to be merged.
        filenames (list, optional): the list of the yaml files to be merged.
        definitions (str or dict, optional): name of the section containing string template
                                                definitions or a dictionary with the same

    Keyword Args:
        loglevel (str, optional): the loglevel to be used, default is 'WARNING'

    Returns:
        A dictionary containing the merged contents of all the yaml files.
    """
    yaml = YAML()  # default, if not specified, is 'rt' (round-trip) # noqa F841

    if isinstance(definitions, str):  # if definitions is a string we need to read twice
        yaml_dict = _load_merge(folder_path=folder_path, definitions=None,
                                filenames=filenames, **kwargs)  # Read without definitions
        definitions = yaml_dict.get(definitions)
        yaml_dict = _load_merge(folder_path=folder_path, definitions=definitions,
                                filenames=filenames, **kwargs)  # Read again with definitions
    else:  # if a dictionary or None has been passed for definitions we read only once
        yaml_dict = _load_merge(folder_path=folder_path, definitions=definitions,
                                filenames=filenames, **kwargs)

    return yaml_dict


def load_yaml(infile: str, definitions: str | dict | None = None, jinja: bool = True):
    """
    Load yaml file with template substitution

    Args:
        infile (str): a file path to the yaml
        definitions (str or dict, optional): name of the section containing string template
                                             definitions or a dictionary with the same content
        jinja: (bool): jinja2 templating is used instead of standard python templating. Default is true.
    Returns:
        A dictionary with the yaml file keys
    """

    if not os.path.exists(infile):
        raise FileNotFoundError(f'ERROR: {infile} not found: you need to have this configuration file!')

    yaml = YAML(typ='rt')  # default, if not specified, is 'rt' (round-trip)

    cfg = None
    # Load the YAML file as a text string
    with open(infile, 'r', encoding='utf-8') as file:
        yaml_text = file.read()

    if isinstance(definitions, str):  # if it is a string extract from original yaml, else it is directly a dict
        cfg = yaml.load(yaml_text)
        definitions = cfg.get(definitions)

    if definitions:
        # perform template substitution with jinja
        if jinja:
            template = Template(yaml_text)
            rendered_yaml = template.render(definitions)
            cfg = yaml.load(rendered_yaml)
        # use default python templating
        else:
            template = DefaultTemplate(yaml_text).safe_substitute(definitions)
            cfg = yaml.load(template)
    else:
        if not cfg:  # did we already load it ?
            cfg = yaml.load(yaml_text)

    return cfg


def dump_yaml(outfile=None, cfg=None, typ='rt'):
    """
    Dump to a custom yaml file

    Args:
        outfile(str):   a file path
        cfg(dict):      a dictionary to be dumped
        typ(str):       the type of YAML initialisation.
                        Default is 'rt' (round-trip)
    """
    # Initialize YAML object
    yaml = YAML(typ=typ)

    yaml.representer.add_representer(
        type(None),
        lambda self, _: self.represent_scalar('tag:yaml.org,2002:null', 'null')
    )

    # Check input
    if outfile is None:
        raise ValueError('ERROR: outfile not defined')
    if cfg is None:
        raise ValueError('ERROR: cfg not defined')

    # Dump the dictionary with a safe temporary directory
    # to avoid intake reading a partially written file

    # Ensure parent directory exists
    dest_dir = os.path.dirname(os.path.abspath(outfile))
    if dest_dir:  # Handle edge case where filename has no directory component
        os.makedirs(dest_dir, exist_ok=True)
    else:
        dest_dir = '.'  # Use current directory
    with TemporaryDirectory(dir=dest_dir) as tmpdirname:
        tmp_file = os.path.join(tmpdirname, "temp.yaml")
        with open(tmp_file, 'w', encoding='utf-8') as file:
            yaml.dump(cfg, file)
        os.replace(tmp_file, outfile)


def _load_merge(folder_path: str | None = None, filenames: list | None = None,
                definitions: str | dict | None = None, merged_dict: dict | None = None,
                loglevel: str = 'WARNING'):
    """
    Helper function for load_merge_yaml.
    Load and merge yaml files located in a given folder
    or a list of yaml files into a dictionary.

    Args:
        folder_path (str, optional):         the path of the folder containing the yaml
                                             files to be merged.
        filenames (list, optional):          the list of the yaml files to be merged.
        definitions (str or dict, optional): name of the section containing string template
                                             definitions or a dictionary with the same content
        merged_dict (dict, optional):        the dictionary to be updated with the yaml files
        loglevel (str, optional):            the loglevel to be used, default is 'WARNING'

    Returns:
        A dictionary containing the merged contents of all the yaml files.

    Raises:
        ValueError: if both folder_path and filenames are None or if both are not None.
    """
    logger = log_configure(log_name='yaml', log_level=loglevel)

    if merged_dict is None:
        logger.debug('Creating a new dictionary')
        merged_dict = defaultdict(dict)
    else:
        logger.debug('Updating an existing dictionary')

    if filenames is None and folder_path is None:
        raise ValueError('ERROR: at least one between folder_path or filenames must be provided')

    if filenames:  # Merging a list of files
        logger.debug(f'Files to be merged: {filenames}')
        for filename in filenames:
            yaml_dict = load_yaml(filename, definitions)
            for key, value in yaml_dict.items():
                merged_dict[key].update(value)

    if folder_path:  # Merging all the files in a folder
        logger.debug(f'Folder to be merged: {folder_path}')
        if not os.path.exists(folder_path):
            raise FileNotFoundError(f'ERROR: {folder_path} not found: it is required to have this folder!')
        for filename in os.listdir(folder_path):
            if filename.endswith(('.yml', '.yaml')):
                file_path = os.path.join(folder_path, filename)
                yaml_dict = load_yaml(file_path, definitions)
                for key, value in yaml_dict.items():
                    merged_dict[key].update(value)

    logger.debug('Dictionary updated')
    logger.debug(f'Keys: {merged_dict.keys()}')

    return merged_dict



import pytest
import numpy as np
from aqua import Reader
from aqua.core.exceptions import NoDataError
from conftest import LOGLEVEL


@pytest.mark.aqua
class TestFilterKey:
    """Tests for the filter_key functionality in AQUA Reader"""
    def test_filter_key_year(self):
        reader = Reader(model="FESOM", exp="test-pi", source="original_2d_filter", 
                        startdate="1985-01-01", enddate="1985-12-31", LOGLEVEL=LOGLEVEL)
        data = reader.retrieve()
        assert data.time.size == 1  # Expecting 12 months of data for the year 1985
        assert np.all(data.time.dt.year == 1985)

    def test_filter_key_year_no_dates(self):
        with pytest.raises(NoDataError, match="No files found after filtering the catalog!"):
            _ = Reader(model="FESOM", exp="test-pi", source="original_2d_filter", 
                        LOGLEVEL=LOGLEVEL, startdate="2020-01-01", enddate="2020-12-31")

    def test_no_filter_key(self):
        reader = Reader(model="FESOM", exp="test-pi", source="original_2d_filter", 
                        LOGLEVEL=LOGLEVEL)
        data = reader.retrieve()
        assert data.time.size == 2  # Expecting multiple years of data


import pytest
import xarray as xr
import numpy as np
import healpy as hp
from scipy.interpolate import griddata
from pypdf import PdfReader

from aqua import Reader
from aqua.core.util.graphics import add_cyclic_lon, plot_box, minmax_maps
from aqua.core.util import cbar_get_label, evaluate_colorbar_limits, add_pdf_metadata
from aqua.core.util import get_nside, get_npix, healpix_resample
from aqua.core.util import coord_names, set_map_title
from aqua.core.graphics import plot_single_map
from conftest import DPI, LOGLEVEL

loglevel = LOGLEVEL


@pytest.fixture()
def da():
    """Create a test DataArray"""
    lon_values = np.linspace(0, 360, 36)  # Longitude values
    lat_values = np.linspace(-90, 90, 18)  # Latitude values
    data = np.random.rand(18, 36)  # Example data
    return xr.DataArray(data, dims=['lat', 'lon'], coords={'lon': lon_values, 'lat': lat_values})


@pytest.mark.graphics
def test_add_cyclic_lon(da):
    """Test the add_cyclic_lon function"""
    old_da = da.copy()
    new_da = add_cyclic_lon(da)

    # Assertions to test the function
    assert isinstance(new_da, xr.DataArray), "Output should be an xarray.DataArray"
    assert 'lon' in new_da.coords, "Output should have a 'lon' coordinate"
    assert 'lat' in new_da.coords, "Output should have a 'lat' coordinate"
    assert np.allclose(new_da.lat, old_da.lat), "Latitude values should be equal"
    assert np.allclose(new_da.isel(lon=-1).values, old_da.isel(lon=0).values), \
           "First and last longitude values should be equal"
    assert new_da.shape == (18, 37), "Output shape is incorrect"

    with pytest.raises(ValueError):
        add_cyclic_lon(da='test')  # Test with invalid input


@pytest.mark.graphics
def test_plot_box():
    """Test the plot box function"""
    num_rows, num_cols = plot_box(10)
    assert num_rows == 3, "Number of rows should be 3"
    assert num_cols == 4, "Number of columns should be 4"

    num_rows, num_cols = plot_box(1)
    assert num_rows == 1, "Number of rows should be 1"
    assert num_cols == 1, "Number of columns should be 1"

    num_rows, num_cols = plot_box(3)
    assert num_rows == 2, "Number of rows should be 2"
    assert num_cols == 2, "Number of columns should be 2"

    with pytest.raises(ValueError):
        plot_box(0)


@pytest.mark.graphics
def test_minmax_maps(da):
    """Test the minmax_maps function"""
    # Create a list of DataArrays
    maps = [da, da + 1, da + 2]

    # Test the function
    min_val, max_val = minmax_maps(maps)

    assert min_val < max_val, "Minimum value should be less than maximum value"
    for i in range(len(maps)):
        assert min_val <= maps[i].min().values, "Minimum value should be less than minimum value of the map"
        assert max_val >= maps[i].max().values, "Maximum value should be greater than maximum value of the map"

@pytest.fixture(scope='module')
def ifs_data():
    """Retrieve IFS data for graphics tools tests"""
    reader = Reader(model='IFS', exp='test-tco79', source='short', loglevel=loglevel)
    return reader.retrieve()

@pytest.mark.graphics
def test_label(ifs_data):
    """Test the cbar_get_label function"""
    da = ifs_data['2t']

    # Test cbar_get_label function
    label = cbar_get_label(da, loglevel=loglevel)
    # assert label is a string
    assert isinstance(label, str), "Colorbar label should be a string"
    assert label == "2 metre temperature [K]", "Colorbar label is incorrect"

    # Test the function with a custom label
    label = cbar_get_label(da, cbar_label='Temperature', loglevel=loglevel)
    assert label == 'Temperature', "Colorbar label is incorrect"

    # Test the cbar limits function with sym=False
    vmin, vmax = evaluate_colorbar_limits(da, sym=False)
    assert vmin < vmax, "Minimum value should be less than maximum value"
    assert vmin == 232.79393005371094, "Minimum value is incorrect"
    assert vmax == 310.61033630371094, "Maximum value is incorrect"

    # Test the cbar limits function with sym=True
    vmin, vmax = evaluate_colorbar_limits(da, sym=True)
    assert vmin < vmax, "Minimum value should be less than maximum value"
    assert vmin == -310.61033630371094, "Minimum value is incorrect"
    assert vmax == 310.61033630371094, "Maximum value is incorrect"

    with pytest.raises(ValueError):
        evaluate_colorbar_limits(maps=None)


@pytest.mark.graphics
def test_pdf_metadata(tmp_path):
    """Test the add_pdf_metadata function"""
    # Generate a test figure from a random xarray DataArray
    da = xr.DataArray(np.random.rand(18, 36), dims=['lat', 'lon'], coords={'lon': np.linspace(0, 360, 36),
                                                                           'lat': np.linspace(-90, 90, 18)})
    fig, _ = plot_single_map(da, title='Test', filename='test', format='pdf',
                             return_fig=True, loglevel=loglevel)

    fig.savefig(tmp_path / 'test.pdf', dpi=DPI)
    filename = str(tmp_path / 'test.pdf')
    # Test the function
    add_pdf_metadata(filename=filename, metadata_value='Test',
                     metadata_name='/Test description', loglevel=loglevel)
    add_pdf_metadata(filename=filename, metadata_value='Test caption',
                     loglevel=loglevel)

    # Open the PDF and check the metadata
    pdf_reader = PdfReader(filename)
    metadata = pdf_reader.metadata

    assert metadata['/Test description'] == 'Test', "Old metadata should be kept"
    assert metadata['/Description'] == 'Test caption', "Description should be added to metadata"


@pytest.mark.graphics
def test_set_map_title(da):
    title = set_map_title(da)

    assert title is None, "Title should be None"


@pytest.mark.graphics
def test_coord_names():
    """Test the coord_names function"""
    # Create a test DataArray
    lon_values = np.linspace(0, 360, 36)
    lat_values = np.linspace(-90, 90, 18)
    data = np.random.rand(18, 36)
    da = xr.DataArray(data, dims=['latitude', 'longitude'],
                      coords={'longitude': lon_values, 'latitude': lat_values})

    # Test the function
    lon_name, lat_name = coord_names(da)
    assert lon_name == 'longitude', "Longitude name is incorrect"
    assert lat_name == 'latitude', "Latitude name is incorrect"

@pytest.mark.graphics
class TestHealpixUtils:
    def test_get_nside_valid_ndarray(self):
        nside = 8
        npix = hp.nside2npix(nside)
        data = np.arange(npix)
        assert get_nside(data) == nside

    def test_get_nside_valid_xarray(self):
        nside = 4
        npix = hp.nside2npix(nside)
        data = xr.DataArray(np.arange(npix))
        assert get_nside(data) == nside

    def test_get_nside_invalid_type(self):
        with pytest.raises(ValueError, match="Input data must be a numpy array or xarray DataArray"):
            get_nside("invalid")

    def test_get_nside_empty_array(self):
        with pytest.raises(ValueError, match="data array is empty"):
            get_nside(np.array([]))

    def test_get_nside_invalid_npix(self):
        data = np.arange(123)  # 123 not a valid npix
        with pytest.raises(ValueError, match="Invalid HEALPix map: npix=123"):
            get_nside(data)

    def test_get_npix_valid(self):
        nside = 16
        data = np.arange(hp.nside2npix(nside))
        expected = hp.nside2npix(nside)
        assert get_npix(data) == expected

    def test_get_npix_invalid_type(self):
        with pytest.raises(ValueError, match="Input data must be a numpy array or xarray DataArray"):
            get_npix("invalid")

    def test_get_npix_invalid_npix(self):
        data = np.arange(999)  # Not valid npix
        with pytest.raises(ValueError, match="Invalid HEALPix map: npix=999"):
            get_npix(data)

    def test_get_npix_empty_array(self):
        with pytest.raises(ValueError, match="data array is empty"):
            get_npix(np.array([]))


@pytest.fixture
def healpix_data():
    nside = 8
    npix = hp.nside2npix(nside)
    data = xr.DataArray(np.random.rand(npix))
    xlims = (-30, 30)
    ylims = (-30, 30)
    return data, xlims, ylims

@pytest.mark.graphics
class TestHealpixResample:
    def test_healpix_resample_nearest(self, healpix_data):
        data, xlims, ylims = healpix_data
        result = healpix_resample(data, xlims=xlims, ylims=ylims, nx=10, ny=10, method="nearest")
        assert isinstance(result, xr.DataArray)
        assert result.ndim == 2
        assert result.shape == (10, 10)

    def test_healpix_resample_linear(self, healpix_data):
        data, xlims, ylims = healpix_data
        result = healpix_resample(data, xlims=xlims, ylims=ylims, nx=10, ny=10, method="linear")
        assert isinstance(result, xr.DataArray)
        assert result.ndim == 2
        assert result.shape == (10, 10)

    def test_healpix_resample_default_grid_size(self, healpix_data):
        data, _, _ = healpix_data
        result = healpix_resample(data, method="nearest")
        assert isinstance(result, xr.DataArray)
        assert result.ndim == 2
        assert result.shape[0] > 0 and result.shape[1] > 0

    def test_healpix_resample_sparse_valid(self):
        nside = 8
        full_npix = hp.nside2npix(nside)  # 768, valid
        selected_cells = np.random.choice(full_npix, size=100, replace=False)
        
        # Build a *sparse* DataArray, but with npix = 768 (valid)
        # Set all values to NaN, except for the selected_cells
        data_array = np.full(full_npix, np.nan, dtype=np.float32)
        values = np.random.rand(100).astype(np.float32)
        data_array[selected_cells] = values
        
        # Assign the 'cell' coordinate
        var = xr.DataArray(data_array, coords={"cell": np.arange(full_npix)}, dims=["cell"])

        result = healpix_resample(var, nx=20, ny=20, method="nearest")
        assert isinstance(result, xr.DataArray)
        assert result.ndim == 2


import matplotlib.pyplot as plt
import xarray as xr
import numpy as np
import pandas as pd
import seaborn as sns
from metpy.units import units
from aqua.core.util import to_list, unit_to_latex
from aqua.core.logger import log_configure
from .styles import ConfigStyle
from matplotlib import colors as mcolors
import textwrap

def boxplot(fldmeans: list[xr.Dataset],
            model_names: list[str],
            variables: list[str],
            variable_names: list[str] = None,
            add_mean_line: bool = True,
            title: str = None,
            style: str = None,
            loglevel: str = 'WARNING'):
    """
    Generate a boxplot of precomputed field-mean values for multiple variables and models.
    A dashed horizontal line is added to indicate the mean for each box (slightly darker than box color).

    Args:
        fldmeans (list of xarray.Dataset): Precomputed fldmean() for each model.
        model_names (list of str): Names corresponding to each fldmean dataset.
        variables (list of str): Variable names to be plotted (as in the fldmean Datasets).
        variable_names (list of str, optional): Display names for the variables.
        add_mean_line (bool, optional): Whether to add dashed lines for means.
        title (str, optional): Title for the plot.
        style (str, optional): Style to apply to the plot.
        loglevel (str): Logging level, unused but kept for compatibility.

    Returns:
        tuple: Matplotlib figure and axis.
    """

    logger = log_configure(loglevel, 'boxplot')
    ConfigStyle(style=style, loglevel=loglevel)
    sns.set_palette("pastel")
    fontsize = 18

    fldmeans, model_names, variables = to_list(fldmeans), to_list(model_names), to_list(variables)

    # Map internal variable names to display names
    if variable_names and len(variable_names) == len(variables):
        labels = dict(zip(variables, variable_names))
    else:
        labels = {v: v for v in variables}

    records = []
    unit_sets = {}

    for ds, model in zip(fldmeans, model_names):
        for var in variables:
            invert = var.startswith('-')
            var_name = var.lstrip('-')

            if var_name not in ds:
                logger.warning(f"Variable '{var_name}' missing in {model}, skipping.")
                continue

            data = ds[var_name].values.flatten()
            data = -data if invert else data
            data = data[np.isfinite(data)]

            # Track units
            unit_attr = ds[var_name].attrs.get('units')
            if unit_attr:
                base = units(unit_attr).to_base_units()
                unit_sets.setdefault(var_name, set()).add(base)

            for val in data:
                records.append({
                    'Variables': labels[var],
                    'Values': val,
                    'Models': model
                })

    df = pd.DataFrame.from_records(records)

    # Create figure
    fig, ax = plt.subplots(figsize=(12, 8))

    # Explicit order for variables and models
    order = [labels[v] for v in variables]
    hue_order = list(model_names)

    # Plot the boxplots
    sns.boxplot(
        data=df, x='Variables', y='Values', hue='Models',
        order=order, hue_order=hue_order, width=0.8, ax=ax
    )

    wrapped_labels = [textwrap.fill(lbl, 14) for lbl in order]
    ax.set_xticks(range(len(wrapped_labels)))
    ax.set_xticklabels(wrapped_labels, fontsize=fontsize)

    # --- Add dashed mean lines for each box ---
    if add_mean_line:
        means = df.groupby(['Variables', 'Models'])['Values'].mean().unstack(fill_value=np.nan)

        n_vars = len(order)
        n_hues = len(hue_order)
        total_box_width = 0.8  # same width passed to sns.boxplot
        if n_hues > 0:
            single_box_width = total_box_width / n_hues
        else:
            single_box_width = total_box_width
        # Offsets for multiple models per variable
        offsets = np.linspace(
            -total_box_width/2 + single_box_width/2,
            total_box_width/2 - single_box_width/2,
            n_hues
        )

        palette = sns.color_palette()

        for i, var in enumerate(order):
            for j, model in enumerate(hue_order):
                if model not in means.columns or var not in means.index:
                    continue
                mean_val = means.loc[var, model]
                if np.isnan(mean_val):
                    continue
                x_center = i + offsets[j]
                x_left = x_center - single_box_width / 2
                x_right = x_center + single_box_width / 2

                # Darken the base color inline
                base_color = palette[j % len(palette)]
                rgb = np.array(mcolors.to_rgb(base_color))
                darker = tuple(rgb * 0.7)  # 0.7 = 30% darker

                ax.hlines(
                    mean_val, x_left, x_right,
                    colors=[darker],
                    linestyles='--', linewidth=2.5, zorder=5
                )

    # Title and labels
    if title:
        ax.set_title(title, fontsize=fontsize + 3)
    else:
        vars_str = ', '.join(labels[v] for v in variables)
        models_str = ', '.join(model_names)
        wrapped_title = textwrap.fill(f'Boxplot of {vars_str}', 70)
        ax.set_title(wrapped_title, fontsize=fontsize+2)
 
    ax.set_xlabel('Variables', fontsize=fontsize)

    # Y-axis label from units
    global_units_found = set().union(*unit_sets.values())
    if len(global_units_found) == 1:
        first_var = variables[0][1:] if variables[0].startswith('-') else variables[0]
        units_str = fldmeans[0][first_var].attrs.get('units', '')
        units_latex = unit_to_latex(units_str) if units_str else ''
        ax.set_ylabel(units_latex, fontsize=fontsize, labelpad=12)    
    else:
        ax.set_ylabel('Values (various units)', fontsize=fontsize, labelpad=12)

    ax.tick_params(axis='x', labelsize=fontsize - 4)
    ax.tick_params(axis='y', labelsize=fontsize - 4)
    ax.legend(loc='upper right', fontsize=fontsize)
    ax.grid(True)
    fig.tight_layout()
    return fig, ax


"""This module base class for grid type builders and its extensions."""
import os
from typing import Optional, Dict
import numpy as np
import xarray as xr
from cdo import Cdo
from smmregrid import CdoGenerate, Regridder, GridType
from aqua.core.logger import log_configure


class BaseGridBuilder:
    """
    Base class for grid type builders.
    """
    requires_bounds = False
    bounds_error_message = "Data has no bounds, cannot create grid"
    logger_name = "BaseGridBuilder"
    CDOZIP = "-f nc4 -z zip"

    def __init__(
        self,
        vert_coord: str,
        original_resolution: str,
        model_name: str,
        grid_name: Optional[str] = None,
        loglevel: str = 'warning'
    ) -> None:
        """
        Initialize the BaseGridBuilder.

        Args:
            vert_coord (str): The vertical coordinate if applicable.
            original_resolution (str): The original resolution of the data.
            model_name (str): The name of the model.
            grid_name (Optional[str]): The name of the grid, if applicable.
            loglevel (str, optional): The logging level for the logger. Defaults to 'warning'.
        """
        self.masked = None
        self.vert_coord = vert_coord
        self.loglevel = loglevel
        self.original_resolution = original_resolution
        self.model_name = model_name
        self.grid_name = grid_name
        self.logger = log_configure(log_level=loglevel, log_name=self.logger_name)
        self.cdo = Cdo()

    def clean_attributes(self, data: xr.Dataset) -> xr.Dataset:
        """
        Clean the attributes of the data.

        Args:
            data (xarray.Dataset): The dataset to clean attributes for.
        Returns:
            xarray.Dataset: The dataset with cleaned attributes.
        """
        # cleaning attributes for variables
        for var in data.data_vars:
            data[var].attrs = {}

        # setting attributes for mask
        data['mask'].attrs['_FillValue'] = -9999
        data['mask'].attrs['missing_value'] = -9999
        data['mask'].attrs['long_name'] = 'mask'
        data['mask'].attrs['units'] = '1'
        data['mask'].attrs['standard_name'] = 'mask'

        # attribute checks for coordinates
        for coord in data.coords:

            # remove axis which can confuse CDO
            if not self.vert_coord or coord != self.vert_coord:
                self.logger.debug("Removing axis for %s", coord)
                if 'axis' in data[coord].attrs:
                    del data[coord].attrs['axis']

            # remove bounds which can confuse CDO
            if not self.has_bounds(data):
                self.logger.debug("No bounds found for %s", coord)
                if 'bounds' in data[coord].attrs:
                    self.logger.debug("Removing bounds for %s", coord)
                    del data[coord].attrs['bounds']

        # adding vertical properties
        if self.vert_coord:
            data[self.vert_coord].attrs['axis'] = 'Z'

        return data

    def has_bounds(self, data: xr.Dataset) -> bool:
        """
        Check if the data has bounds.

        Args:
            data (xarray.Dataset): The dataset to check for bounds.
        Returns:
            bool: True if bounds are present, False otherwise.
        """
        if 'lon_bounds' in data.variables and 'lat_bounds' in data.variables:
            return True
        if 'lon_bnds' in data.variables and 'lat_bnds' in data.variables:
            return True
        return False

    def get_metadata(self, data: xr.Dataset) -> dict:
        """
        Abstract method to get metadata for the grid type. Must be implemented by subclasses.

        Args:
            data (xarray.Dataset): The dataset to extract metadata from.
        Returns:
            dict: Metadata dictionary for the grid type.
        """
        raise NotImplementedError("Subclasses must implement get_metadata()")

    def data_reduction(self, data: xr.Dataset, gridtype: GridType, vert_coord: Optional[str] = None) -> xr.Dataset:
        """
        Reduce the data to a single variable and time step.
        Args:
            data (xarray.Dataset): The dataset containing grid data.
            gridtype (GridInspector): The grid object containing GridType info.
            vert_coord (str, optional): The vertical coordinate if applicable.
        Returns:
            xarray.Dataset: The reduced data.
        """
        # extract first var from GridType and get the attributes of the original variable
        var = next(iter(gridtype.variables))
        attrs = data[var].attrs.copy()

        # guess time dimension from the GridType
        timedim = gridtype.time_dims[0] if gridtype.time_dims else None

        # temporal reduction
        if timedim:
            data = data.isel({timedim: 0}, drop=True)

        # load the variables and rename to mask for consistency
        space_bounds = [bound for bound in gridtype.bounds if not 'time' in bound]
        load_vars = [var] + space_bounds  # (gridtype.bounds or [])
        data = data[load_vars]
        data = data.rename({var: 'mask'})

        # drop the remnant vertical coordinate if present
        if vert_coord and f"idx_{vert_coord}" in data.coords:
            data = data.drop_vars(f"idx_{vert_coord}")

        # set the mask variable to 1 where data is not null
        data['mask'] = xr.where(data['mask'].isnull(), np.nan, 1)

        # preserve the attributes of the original variable
        data['mask'].attrs = attrs

        return data

    def select_2d_slice(self, data: xr.Dataset, vert_coord: Optional[str] = None) -> xr.Dataset:
        """
        Select a 2D slice from the data along the vertical coordinate, if present.
        Args:
            data (xarray.Dataset): The dataset containing grid data.
            vert_coord (str, optional): The vertical coordinate if applicable.
        Returns:
            xarray.Dataset: The 2D-sliced data.
        """
        if vert_coord and vert_coord in data.dims:
            data2d = data.isel({vert_coord: 0})
        else:
            data2d = data
        if isinstance(data2d, xr.DataArray):
            data2d = data2d.to_dataset()
        return data2d

    def detect_mask_type(self, data: xr.Dataset) -> Optional[str]:
        """
        Detect the type of mask based on the data.
        Returns 'oce', 'land', or None.
        Args:
            data (xarray.Dataset): The dataset containing the 'mask' variable.
        Returns:
            Optional[str]: 'oce', 'land', or None if no mask is detected.
        """
        nan_count = float(data['mask'].isnull().sum().values) / data['mask'].size
        self.logger.info("Nan count: %s", nan_count)
        if nan_count == 0:
            self.masked = None
        elif 0 < nan_count < 0.5:
            self.masked = "ocean"
        elif nan_count >= 0.5:
            self.masked = "land"
        else:
            raise ValueError(f"Unexpected nan count {nan_count}")
        return self.masked

    def verify_weights(
        self, filename: str, metadata: Dict, target_grid: str = "r180x90"
    ) -> None:
        """
        Verify the creation of the weights from the grid file.

        Args:
            filename (str): Path to the grid file. Could be also a CDO grid name.
            metadata (dict): Metadata dictionary for weights generation.
            target_grid (str, optional): Target grid for weights generation. Defaults to "r180x90".
        Returns:
            None
        """
        remap_method = metadata.get('remap_method', "con")
        cdo_options = metadata.get('cdo_options', "")
        try:
            self.logger.info(
                "Generating weights for %s with method %s and vert_coord %s",
                filename,
                remap_method,
                self.vert_coord)
            generator = CdoGenerate(
                source_grid=filename,
                target_grid=target_grid,
                cdo_options=cdo_options,
                loglevel=self.loglevel)
            weights = generator.weights(method=remap_method, vert_coord=self.vert_coord)
            self.logger.info(
                "Weights %s generated successfully for %s!!! This grid file is approved for AQUA, take a bow!",
                remap_method,
                filename)
        except Exception as e:
            self.logger.error("Error generating weights, something is wrong with weights generation: %s", e)
            raise
        try:
            regridder = Regridder(weights=weights, cdo_options=cdo_options, loglevel=self.loglevel)
            if os.path.exists(filename):
                data = xr.open_dataset(filename)
            else:
                data = self.cdo.const(f'1,{filename}', options=cdo_options, returnXDataset=True)
            regridder.regrid(data)
            self.logger.info(
                "Grid %s regridded successfully for %s!!! This grid file is approved for AQUA, fly me to the moon!",
                remap_method,
                filename)
        except Exception as e:
            self.logger.error("Error regridding, something is wrong with the regridding: %s", e)
            raise

    def write_gridfile(self, input_file: str, output_file: str, metadata: dict) -> None:
        """
        Write the grid file using CDO or by copying, depending on grid type.
        Can be overridden by subclasses for custom behavior.
        Args:
            input_file (str): Path to the temporary input file.
            metadata (dict): Metadata dictionary from prepare().
            output_file (str): Path to the final output file.
        Returns:
            None
        """
        if metadata.get('cdogrid'):
            self.logger.info("Writing grid file to %s with CDO grid %s", output_file, metadata['cdogrid'])
            self.cdo.setgrid(metadata['cdogrid'], input=input_file, output=output_file, options=self.CDOZIP)
        else:
            self.logger.info("Writing grid file to %s", output_file)
            self.cdo.copy(input=input_file, output=output_file, options=self.CDOZIP)



from matplotlib import rcParams
import matplotlib.pyplot as plt
import xarray as xr
from aqua.core.logger import log_configure
from aqua.core.util import evaluate_colorbar_limits, to_list, unit_to_latex
from .styles import ConfigStyle


def plot_gregory_monthly(t2m_monthly_data, net_toa_monthly_data,
                         t2m_monthly_ref: xr.DataArray = None,
                         net_toa_monthly_ref: xr.DataArray = None,
                         fig: plt.Figure = None, ax: plt.Axes = None,
                         set_axis_limits: bool = True,
                         labels: list = None, ref_label: str = None,
                         xlabel: str = None,
                         ylabel: str = None,
                         title: str = 'Monthly Mean',
                         style: str = None, loglevel: str = 'WARNING'):
    """"
    Plot a Gregory plot for monthly data.

    Args:
        t2m_monthly_data (list): List of 2 m temperature data for each month.
        net_toa_monthly_data (list): List of net radiation TOA data for each month.
        t2m_monthly_ref (xr.DataArray, optional): Reference 2 m temperature data.
        net_toa_monthly_ref (xr.DataArray, optional): Reference net radiation TOA data.
        fig (plt.Figure, optional): Figure object to plot on.
        ax (plt.Axes, optional): Axes object to plot on.
        set_axis_limits (bool, optional): Whether to set axis limits. Defaults to True.
        labels (list, optional): List of labels for each month.
        ref_label (str, optional): Label for the reference data.
        title (str, optional): Title of the plot. Not used if None
        style (str, optional): Style for the plot. Defaults is the AQUA default style.
        loglevel (str, optional): Log level for logging. Defaults to 'WARNING'.

    Returns:
        tuple: Figure and Axes objects.
    """
    logger = log_configure(loglevel, 'plot_gregory_monthly')
    ConfigStyle(style=style, loglevel=loglevel)
    rcParams['text.usetex'] = False  # Disable LaTeX rendering for speed

    # We load the data for speed
    t2m_monthly_data = to_list(t2m_monthly_data)
    net_toa_monthly_data = to_list(net_toa_monthly_data)
    t2m_monthly_data = [t2m_monthly_data[i].load() for i in range(len(t2m_monthly_data))]
    net_toa_monthly_data = [net_toa_monthly_data[i].load() for i in range(len(net_toa_monthly_data))]
    t2m_monthly_ref = t2m_monthly_ref.load() if t2m_monthly_ref is not None else None
    net_toa_monthly_ref = net_toa_monthly_ref.load() if net_toa_monthly_ref is not None else None

    labels = to_list(labels) if labels else [None for _ in range(len(t2m_monthly_data))]

    if fig is None and ax is None:
        logger.debug("Creating new figure and axis")
        fig, ax = plt.subplots(1, 1, figsize=(6, 6))

    if xlabel is None:
        xlabel = '2 m Temperature [°C]'
    if ylabel is None:
        ylabel = r"Net radiation TOA [W m$^{-2}$]"

    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.axhline(0, color="k")
    ax.grid(True)

    ref = t2m_monthly_ref is not None and net_toa_monthly_ref is not None

    # Create a cycle that is the average of the available ref data
    if ref:
        t2m_ref = t2m_monthly_ref.groupby('time.month').mean(dim='time')
        net_toa_ref = net_toa_monthly_ref.groupby('time.month').mean(dim='time')
        # Add an extra point same as the first one to close the loop
        t2m_ref = xr.concat([t2m_ref, t2m_ref.isel(month=0)], dim='month', coords='different', compat='equals')
        net_toa_ref = xr.concat([net_toa_ref, net_toa_ref.isel(month=0)], dim='month', coords='different', compat='equals')

    if set_axis_limits:
        # We set a fixed x and y range but then we expand it if data
        # goes beyond the limits
        t2m_list = t2m_monthly_data + to_list(t2m_ref) if ref else t2m_monthly_data
        t2m_min, t2m_max = evaluate_colorbar_limits(t2m_list, sym=False)
        t2m_min = min(t2m_min, min(t2m_ref.values)) if ref else 11.5
        t2m_max = max(t2m_max, max(t2m_ref.values)) if ref else 16.5
        t2m_min = min(t2m_min, 11.5)
        t2m_min = t2m_min - 0.5
        t2m_max = max(t2m_max, 16.5)
        t2m_max = t2m_max + 0.5

        net_toa_list = net_toa_monthly_data + to_list(net_toa_ref) if ref else net_toa_monthly_data
        toa_min, toa_max = evaluate_colorbar_limits(net_toa_list, sym=False)
        toa_min = min(toa_min, min(net_toa_ref.values)) if ref else -11.5
        toa_max = max(toa_max, max(net_toa_ref.values)) if ref else 11.5
        toa_min = min(toa_min, -11.5)
        toa_min = toa_min - 0.5
        toa_max = max(toa_max, 11.5)
        toa_max = toa_max + 0.5

        ax.set_xbound(t2m_min, t2m_max)
        ax.set_ybound(toa_min, toa_max)
        ax.set_xlim(t2m_min, t2m_max)
        ax.set_ylim(toa_min, toa_max)

        logger.debug(f"Monthly x-axis limits: {t2m_min} to {t2m_max}")
        logger.debug(f"Monthly y-axis limits: {toa_min} to {toa_max}")

    for i, (t2m_monthly, net_toa_monthly) in enumerate(zip(t2m_monthly_data, net_toa_monthly_data)):
        ax.plot(t2m_monthly, net_toa_monthly, label=labels[i], marker='o')
    if ref:
        ax.plot(t2m_ref, net_toa_ref, label=ref_label, marker='o', color='black', zorder=3)
        ax.scatter(t2m_ref, net_toa_ref, color='black', s=150, zorder=3)

        # Optimized text rendering
        for m, x, y in zip(range(1, 13), t2m_ref.values[:-1], net_toa_ref.values[:-1]):
            ax.annotate(str(m), (x, y), color='white', fontsize=8, ha='center',
                        va='center', fontweight='bold', zorder=4)

    return fig, ax


def plot_gregory_annual(t2m_annual_data, net_toa_annual_data,
                        t2m_annual_ref: xr.DataArray = None,
                        net_toa_annual_ref: xr.DataArray = None,
                        t2m_std: xr.DataArray = None,
                        net_toa_std: xr.DataArray = None,
                        fig: plt.Figure = None, ax: plt.Axes = None,
                        set_axis_limits: bool = True,
                        labels: list = None,
                        xlabel: str = None,
                        ylabel: str = None,
                        title: str = 'Annual Mean',
                        style: str = None, loglevel: str = 'WARNING'):
    """
    Plot a Gregory plot for annual data.

    Args:
        t2m_annual_data (list): List of 2 m temperature data for each year.
        net_toa_annual_data (list): List of net radiation TOA data for each year.
        t2m_annual_ref (xr.DataArray, optional): Reference 2 m temperature data.
        net_toa_annual_ref (xr.DataArray, optional): Reference net radiation TOA data.
        t2m_std (xr.DataArray, optional): Standard deviation of 2 m temperature data.
        net_toa_std (xr.DataArray, optional): Standard deviation of net radiation TOA data.
        fig (plt.Figure, optional): Figure object to plot on.
        ax (plt.Axes, optional): Axes object to plot on.
        set_axis_limits (bool, optional): Whether to set axis limits. Defaults to True.
        labels (list, optional): List of labels for each year.
        ref_label (str, optional): Label for the reference data.
        xlabel (str, optional): Title of the x-axis. Defaults to '2 m Temperature [°C]'.
        ylabel (str, optional): Title of the y-axis. Defaults to "Net radiation TOA [W/m^2]".
        title (str, optional): Title of the plot. Not used if None
        style (str, optional): Style for the plot. Defaults is the AQUA default style.
        loglevel (str, optional): Log level for logging. Defaults to 'WARNING'.

    Returns:
        tuple: Figure and Axes objects.
    """
    logger = log_configure(loglevel, 'plot_gregory_annual')
    ConfigStyle(style=style, loglevel=loglevel)

    labels = to_list(labels) if labels else [None for _ in range(len(t2m_annual_data))]

    # We load the data for speed
    t2m_annual_data = to_list(t2m_annual_data)
    net_toa_annual_data = to_list(net_toa_annual_data)
    t2m_annual_data = [t2m_annual_data[i].load() for i in range(len(t2m_annual_data))]
    net_toa_annual_data = [net_toa_annual_data[i].load() for i in range(len(net_toa_annual_data))]
    t2m_annual_ref = t2m_annual_ref.load() if t2m_annual_ref is not None else None
    net_toa_annual_ref = net_toa_annual_ref.load() if net_toa_annual_ref is not None else None
    t2m_std = t2m_std.load() if t2m_std is not None else None
    net_toa_std = net_toa_std.load() if net_toa_std is not None else None

    if fig is None and ax is None:
        logger.debug("Creating new figure and axis")
        fig, ax = plt.subplots(1, 1, figsize=(6, 6))

    if xlabel is None:
        xlabel = '2 m Temperature [°C]'
    if ylabel is None:
        ylabel = r"Net radiation TOA [W m$^{-2}$]"

    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.axhline(0, color="k")
    ax.grid(True)

    ref = t2m_std is not None and net_toa_std is not None

    if set_axis_limits:
        # We set a fixed x and y range but then we expand it if data
        # goes beyond the limits
        t2m_min, t2m_max = evaluate_colorbar_limits(t2m_annual_data, sym=False)
        t2m_min = min(t2m_min, 13.5)
        t2m_min = t2m_min - 0.5
        t2m_max = max(t2m_max, 14.5)
        t2m_max = t2m_max + 0.5

        toa_min, toa_max = evaluate_colorbar_limits(net_toa_annual_data, sym=False)
        toa_min = min(toa_min, -1.5)
        toa_min = toa_min - 0.5
        toa_max = max(toa_max, 1.5)
        toa_max = toa_max + 0.5

        ax.set_xbound(t2m_min, t2m_max)
        ax.set_ybound(toa_min, toa_max)
        ax.set_xlim(t2m_min, t2m_max)
        ax.set_ylim(toa_min, toa_max)

        logger.debug(f"Annual x-axis limits: {t2m_min} to {t2m_max}")
        logger.debug(f"Annual y-axis limits: {toa_min} to {toa_max}")

    for i, (t2m_annual, net_toa_annual) in enumerate(zip(t2m_annual_data, net_toa_annual_data)):
        ax.plot(t2m_annual, net_toa_annual, label=labels[i], marker='o')

        # We plot the first and last points with different markers
        ax.plot(t2m_annual[0], net_toa_annual[0], marker=">", color="tab:green")
        ax.plot(t2m_annual[-1], net_toa_annual[-1], marker="<", color="tab:red")
        ax.annotate(str(t2m_annual.time.dt.year[0].values), (t2m_annual[0], net_toa_annual[0]),
                    fontsize=8, ha='right')
        ax.annotate(str(t2m_annual.time.dt.year[-1].values), (t2m_annual[-1], net_toa_annual[-1]),
                    fontsize=8, ha='right')
    if ref:
        t2m_mean = t2m_annual_ref.mean(dim='time')
        net_toa_mean = net_toa_annual_ref.mean(dim='time')
        ax.axhspan(net_toa_mean - net_toa_std, net_toa_mean + net_toa_std,
                   color="lightgreen", alpha=0.3, label=r"1 $\sigma$ band")
        ax.axvspan(t2m_mean - t2m_std, t2m_mean + t2m_std,
                   color="lightgreen", alpha=0.3)

    return fig, ax



import xarray as xr
import matplotlib.pyplot as plt
from aqua.core.logger import log_configure
from .styles import ConfigStyle


def index_plot(index: xr.DataArray, thresh: float = 0,
               fig: plt.Figure = None, ax: plt.Axes = None,
               style: str = None, figsize: tuple = (11, 8.5),
               title: str = None, ylim: tuple = None,
               ylabel: str = None, label: str = None, loglevel='WARNING'):
    """
    Index plot together with a black line at index=0.
    Values above thresh are filled in red, values below thresh are filled in blue.

    Args:
        index (DataArray):     Index DataArray
        thresh (float,opt):    Threshold for the index, default is 0
        fig (Figure,opt):      Figure object
        ax (Axes,opt):         Axes object
        style (str, optional): Style to use. Defaults to None (aqua style).
        figsize (tuple,opt):   Figure size, default is (11, 8.5)
        title (str,opt):       Title for the plot. Default is None
        ylim (tuple,opt):      y-axis limits. Default is None
        ylabel (str,opt):      y-axis label. Default is None
        label (str,opt):       Label for the plot. Default is None
        loglevel (str,opt):    Loglevel for the logger. Default is 'WARNING'

    Returns:
        fig, ax: Figure and Axes objects
    """
    logger = log_configure(loglevel, 'index_plot')
    ConfigStyle(style=style, loglevel=loglevel)

    logger.debug('Loading data in memory')
    index = index.load()

    if fig is None:
        fig = plt.figure(figsize=figsize)
    if ax is None:
        ax = fig.add_subplot(111)

    if ylim is not None:
        ax.set_ylim(ylim)

    # Plot the index
    ax.fill_between(x=index.time, y1=thresh, y2=index.values, where=index.values >= thresh,
                    alpha=0.6, color='red', interpolate=True)
    ax.fill_between(index.time, y1=-thresh, y2=index.values, where=index.values < -thresh,
                    alpha=0.6, color='blue', interpolate=True)
    index.plot(ax=ax, color='black', alpha=0.8, label=label)

    ax.axhline(y=0, color='black', alpha=0.5)
    ax.grid(True, axis="y", linestyle='-', color='silver', alpha=0.8)

    if title is not None:
        ax.set_title(title)

    # Set the ylabel
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    else:
        ax.set_ylabel('Index')

    if label is not None:
        ax.legend()

    return fig, ax


def indexes_plot(indexes: list, thresh: float = 0,
                 style=None, figsize: tuple = None,
                 titles: list = None, labels: list = None,
                 suptitle: str = None,
                 ylabel: str = None, loglevel='WARNING'):
    """
    Use the plot_index function to plot two indexes in two
    subplots.

    Args:
        indexes (list):         List of DataArray objects
        thresh (float,opt):     Threshold for the index, default is 0
        style (str, optional):  Style to use. Defaults to None (aqua style).
        figsize (tuple,opt):    Figure size, default is (5.5*n_indexes, 8.5)
        titles (list,opt):      List of strings for the titles of the plots. Default is None
        labels (list,opt):      List of strings for the labels of the plots. Default is None
        suptitle (str,opt):     Title for the figure. Default is None
        ylabel (str,opt):       y-axis label. Default is None
        loglevel (str,opt):     Loglevel for the logger. Default is 'WARNING'

    Returns:
        fig, axs: Figure and Axes objects
    """
    figsize = (8.5, 5.5 * len(indexes)) if figsize is None else figsize
    ConfigStyle(style=style, loglevel=loglevel) 

    fig, axs = plt.subplots(nrows=len(indexes), ncols=1, figsize=figsize, sharex=False)

    # Evaluating a common ylim:
    ymin = min(index.min().values for index in indexes)
    ymax = max(index.max().values for index in indexes)
    ylim = [ymin, ymax]

    if isinstance(axs, plt.Axes):
        axs = [axs]

    for i, index in enumerate(indexes):
        fig, axs[i] = index_plot(index, thresh=thresh,
                                 fig=fig, ax=axs[i],
                                 style=style, figsize=figsize,
                                 title=titles[i] if titles is not None else None,
                                 label=labels[i] if labels is not None else None,
                                 ylim=ylim, ylabel=ylabel,
                                 loglevel=loglevel)

    if suptitle is not None:
        fig.suptitle(suptitle)

    fig.tight_layout()

    return fig, axs



import xarray as xr
import matplotlib.pyplot as plt
import numpy as np

from aqua.core.logger import log_configure
from aqua.core.util import to_list, unit_to_latex
from .styles import ConfigStyle


def plot_histogram(data: xr.DataArray | list[xr.DataArray],
                   ref_data: xr.DataArray | None = None,
                   data_labels: list | None = None,
                   ref_label: str | None = None,
                   style: str | None = None,
                   fig: plt.Figure | None = None,
                   ax: plt.Axes | None = None,
                   figsize: tuple = (10, 6),
                   title: str | None = None,
                   xlogscale: bool = False,
                   ylogscale: bool = True,
                   xmax: float | None = None,
                   xmin: float | None = None,
                   ymax: float | None = None,
                   ymin: float | None = None,
                   smooth: bool = False,
                   smooth_window: int = 5,
                   loglevel: str = 'WARNING'):
    """
    Plot histogram or PDF data.

    Args:
        data (xr.DataArray | list[xr.DataArray]): Histogram data to plot. 
            Must be xarray DataArrays with 'center_of_bin' dimension.
            Can be a single DataArray or a list of DataArrays.
        ref_data (xr.DataArray, optional): Reference histogram data to plot.
        data_labels (list | None, optional): Labels for the data.
        ref_label (str | None, optional): Label for the reference data.
        style (str | None, optional): Style for the plot.
        fig (plt.Figure | None, optional): Matplotlib figure object.
        ax (plt.Axes | None, optional): Matplotlib axes object.
        figsize (tuple, optional): Figure size if a new figure is created.
        title (str | None, optional): Title for the plot.
        xlogscale (bool, optional): Use logarithmic scale for x-axis.
        ylogscale (bool, optional): Use logarithmic scale for y-axis.
        xmax (float | None, optional): Maximum value for x-axis.
        xmin (float | None, optional): Minimum value for x-axis.
        ymax (float | None, optional): Maximum value for y-axis.
        ymin (float | None, optional): Minimum value for y-axis.
        smooth (bool, optional): Apply smoothing to the data.
        smooth_window (int, optional): Window size for smoothing.
        loglevel (str, optional): Logging level.

    Returns:
        tuple: Matplotlib figure and axes objects.
    """
    logger = log_configure(loglevel, 'plot_histogram')
    ConfigStyle(style=style, loglevel=loglevel)
    
    # Convert data to list
    data_list = to_list(data)

    # Create figure if needed
    if fig is None:
        fig = plt.figure(figsize=figsize)
    if ax is None:
        ax = fig.add_subplot(1, 1, 1)

    logger.debug(f"Plotting {len(data_list)} histograms")

    # Plot data
    for i, d in enumerate(data_list):
        if 'center_of_bin' not in d.dims:
            logger.warning(f"Data {i} has no 'center_of_bin' dimension, skipping")
            continue
        
        x = d.center_of_bin.values
        y = d.values
        
        # Apply smoothing if requested
        if smooth:
            y = _smooth_data(y, window_size=smooth_window)
        
        label = data_labels[i] if data_labels and i < len(data_labels) else None
        ax.plot(x, y, label=label, linewidth=2, zorder=3)

    # Handle reference data
    if ref_data is not None:
        if 'center_of_bin' in ref_data.dims:
            x_ref = ref_data.center_of_bin.values
            y_ref = ref_data.values
            
            # Apply smoothing if requested
            if smooth:
                y_ref = _smooth_data(y_ref, window_size=smooth_window)
            
            ax.plot(x_ref, y_ref, 
                   label=ref_label if ref_label else 'Reference',
                   color='black', linestyle='-', linewidth=2, alpha=1.0, zorder=1)

    # Set scales
    if xlogscale:
        ax.set_xscale('log')
    if ylogscale:
        ax.set_yscale('log')

    # Set limits
    if xmin is not None or xmax is not None:
        ax.set_xlim(left=xmin, right=xmax)
    if ymin is not None or ymax is not None:
        ax.set_ylim(bottom=ymin, top=ymax)

    # Finalize plot
    if data_labels or ref_label:
        ax.legend(fontsize='small', loc='upper right')
    
    ax.grid(True, linestyle='-', alpha=0.3)
    
    # Set labels
    first_data = data_list[0]
    xlabel = "Value"
    if hasattr(first_data, 'center_of_bin') and hasattr(first_data.center_of_bin, 'units'):
        xlabel += f" [{unit_to_latex(first_data.center_of_bin.units)}]"
    ax.set_xlabel(xlabel)
    
    ylabel = "Frequency"
    if hasattr(first_data, 'units'):
        ylabel += f" [{unit_to_latex(first_data.units)}]"
    ax.set_ylabel(ylabel)

    # Set title if provided
    if title:
        ax.set_title(title, fontsize=13, fontweight='bold')

    return fig, ax


def _smooth_data(data, window_size=5):
    """
    Apply moving average smoothing to data.
    
    Args:
        data (array): Data to smooth.
        window_size (int): Size of smoothing window.
    
    Returns:
        array: Smoothed data.
    """
    if window_size < 2:
        return data
    
    # Simple moving average
    kernel = np.ones(window_size) / window_size
    smoothed = np.convolve(data, kernel, mode='same')
    
    # Fix edges
    for i in range(window_size // 2):
        smoothed[i] = np.mean(data[:i+window_size//2+1])
        smoothed[-(i+1)] = np.mean(data[-(i+window_size//2+1):])
    
    return smoothed


"""Zarr reference module"""

import os
import json
import xarray as xr
from kerchunk.hdf import SingleHdf5ToZarr
from kerchunk.combine import MultiZarrToZarr
from aqua.core.logger import log_configure


def create_zarr_reference(filelist, outfile, loglevel='WARNING'):
    """
    Create a Zarr file from a list of HDF5/NetCDF files.

    Args:
        filelist (list): A list of file paths to HDF5 files.
        outfile (str): The path to the output Zarr file.
        loglevel (str, optional): The log level for logging. Defaults to 'WARNING'.

    Returns:
        None
    """

    logger = log_configure(log_level=loglevel, log_name='Zarr reference creator')
    data = xr.open_mfdataset(filelist, combine='by_coords')
    identical_coords = [coord for coord in data.coords if coord != 'time']
    logger.debug('Common coordinates: %s', identical_coords)

    logger.debug('Creating Zarr file from %s', filelist)
    singles = [SingleHdf5ToZarr(filepath, inline_threshold=0).translate() for filepath in sorted(filelist)]

    logger.debug('Combining Zarr files')
    mzz = MultiZarrToZarr(
        singles,
        concat_dims=["time"],
        identical_dims=identical_coords,
    )

    logger.debug('Translating Zarr files to json')
    try:
        out = mzz.translate()
    except ValueError as e:
        logger.error('Cannot create Zarr %s file due chunk mismatch', outfile)
        logger.error(e)
        return None

    # Dump to file
    logger.info('Dumping to file JSON %s', outfile)
    if os.path.exists(outfile):
        os.remove(outfile)
    with open(outfile, "w") as file:
        json.dump(out, file)

    return outfile



import xarray as xr
import matplotlib.pyplot as plt

from .styles import ConfigStyle
from .lat_lon_profiles import plot_lat_lon_profiles
            
def plot_seasonal_lat_lon_profiles(seasonal_data,
                                   ref_data=None,
                                   ref_std_data=None,
                                   style: str = None,
                                   loglevel='WARNING',
                                   data_labels: list = None,
                                   title: str = None,
                                   ref_label: str = None
                                   ):
    """
    Plot seasonal lat-lon profiles in a 2x2 subplot layout for the four meteorological seasons.

    This function creates exactly 4 subplots arranged in a 2x2 grid, each showing lat-lon 
    profiles for a specific season. The seasons are hardcoded and must be provided in the 
    exact order: [DJF, MAM, JJA, SON].

    Args:
        seasonal_data (list): List of exactly 4 elements, one for each season.
            Must be in order: [DJF, MAM, JJA, SON].
            Each element can be either:
            - A single xarray DataArray (for single model)
            - A list of xarray DataArrays (for multiple models)
            
            Examples:
            Single model: [djf_data, mam_data, jja_data, son_data]
            Multiple models: [[model1_djf, model2_djf], [model1_mam, model2_mam], ...]
            
            DJF = December-January-February (Winter)
            MAM = March-April-May (Spring) 
            JJA = June-July-August (Summer)
            SON = September-October-November (Autumn)
        ref_data (list, optional): Reference data for each season, same structure as seasonal_data.
        ref_std_data (list, optional): Reference standard deviation data for each season.
        style (str, optional): Style configuration for the plot.
        loglevel (str): Logging level.
        data_labels (list, optional): List of data_labels for each subplot. If provided, must have 4 elements.
        title (str, optional): Overall title for the 2x2 subplot figure.
        ref_label (str, optional): Label for the reference data in the legend.

    Returns:
        fig, axs: Matplotlib figure and axes objects (2x2 subplot layout).
        
    Raises:
        ValueError: If seasonal_data is not a list of exactly 4 elements.
    """
    ConfigStyle(style=style, loglevel=loglevel)
    
    # Validate seasonal_data structure
    if not isinstance(seasonal_data, list) or len(seasonal_data) != 4:
        raise ValueError("seasonal_data must be a list of 4 elements: [DJF, MAM, JJA, SON]")
    
    # Validate and prepare data labels
    for i, season_data in enumerate(seasonal_data):
        if isinstance(season_data, list):
            if not all(isinstance(d, xr.DataArray) for d in season_data):
                raise ValueError(f"Season {i} contains non-DataArray elements")
        elif not isinstance(season_data, xr.DataArray):
            raise ValueError(f"Season {i} must be DataArray or list of DataArrays")

    # Validate ref_std_data if provided
    if ref_std_data is not None:
        computed_ref_std_data = []
        for s in ref_std_data:
            if s is not None and hasattr(s, 'compute'):
                computed_ref_std_data.append(s.compute())
            else:
                computed_ref_std_data.append(s)
        ref_std_data = computed_ref_std_data
    else:
        ref_std_data = [None]

    fig, axs = plt.subplots(2, 2, figsize=(12, 8), constrained_layout=True)
    axs = axs.flatten()
    season_names = ["DJF", "MAM", "JJA", "SON"]

    # Plot the 4 seasonal subplots
    for i, ax in enumerate(axs):
        season_data = seasonal_data[i]
        season_title = season_names[i]
        season_ref_data = ref_data[i] if ref_data is not None and i < len(ref_data) else None
        season_ref_std_data = ref_std_data[i] if ref_std_data is not None and i < len(ref_std_data) else None
        
            
        _, _ = plot_lat_lon_profiles(data=season_data,
                                     ref_data=season_ref_data,
                                     ref_std_data=season_ref_std_data,
                                     ref_label=ref_label,
                                     data_labels=data_labels,
                                     fig=fig, ax=ax,
                                     loglevel=loglevel)

        ax.set_title(season_title)
        ax.grid(True, linestyle='--', alpha=0.7)
        
        # Show legend only on the first subplot
        if i == 0:
            ax.legend(fontsize='small', loc='upper right')
        elif hasattr(ax, 'legend_') and ax.legend_:
            ax.legend_.remove()
    
    # Add overall title if provided
    if title:
        fig.suptitle(title, fontsize=14, fontweight='bold', y=0.98)

    fig.set_layout_engine('tight')

    return fig, axs
    


"""
This module contains the CLI for the GridBuilder.
"""

import argparse
from aqua import Reader
from aqua import GridBuilder
from aqua.core.util import load_yaml, get_arg


def builder_parser(parser=None):

    """
    Parse command line arguments for the builder CLI.
    """
    if parser is None:
        parser = argparse.ArgumentParser(description='AQUA grids builder CLI')
    parser.add_argument('-c', '--config', type=str, 
                        help='YAML configuration file for the builder [default: None]')
    parser.add_argument('--catalog', type=str, help='Catalog for the Reader')
    parser.add_argument('-m', '--model', type=str,
                        help='Model name (e.g. "ECMWF", "ERA5")')
    parser.add_argument('-e', '--exp', type=str,
                        help='Experiment name (e.g. "historical", "future")')
    parser.add_argument('-s', '--source', type=str,
                        help='Data source (e.g. "reanalysis", "forecast")')
    parser.add_argument('-l', '--loglevel', type=str, default=None,
                        help='Log level [default: WARNING]')
    parser.add_argument('--rebuild', action='store_true',
                        help='Rebuild the grid even if it already exists')
    parser.add_argument('--vert_coord', type=str,
                        help='Vertical coordinate name for 3D grids [default: None]')
    parser.add_argument('--version', type=int,
                        help='Version number for the grid file [default: None]')
    parser.add_argument('--outdir', type=str, default='.',
                        help='Output directory for the grid file [default: current directory]')
    parser.add_argument('--original', type=str,
                        help='Original resolution of the grid [default: None]')
    parser.add_argument('--modelname', type=str,
                        help='alternative name for the model for grid naming [default: None]')
    parser.add_argument('--gridname', type=str,
                        help='alternative name for the grid for grid naming [default: None]. Required for Curvilinear and Unstructured grids.')
    parser.add_argument('--fix', action='store_true',
                        help='Fix the original source [default: False]')
    parser.add_argument('--verify', action='store_true', default=False,
                        help='Verify the grid file after creation [default: False]')
    parser.add_argument('--yaml', action='store_true', default=False,
                        help='Create the grid entry in the grid file [default: False]')

    return parser

def builder_execute(args):
    """
    Execute the builder CLI with the provided arguments or configuration file.
    """
    config = {}
    reader_config = {}
    builder_config = {}
    if hasattr(args, 'config') and args.config:
        config = load_yaml(args.config)
        reader_config = config.get('reader', {})
        builder_config = config.get('builder', {})

    # Use get_arg to merge CLI args and config file values
    catalog = get_arg(args, 'catalog', reader_config.get('catalog'))
    model = get_arg(args, 'model', reader_config.get('model'))
    exp = get_arg(args, 'exp', reader_config.get('exp'))
    source = get_arg(args, 'source', reader_config.get('source'))
    fix = get_arg(args, 'fix', reader_config.get('fix', False))
    loglevel = get_arg(args, 'loglevel', builder_config.get('loglevel', 'WARNING'))
    outdir = get_arg(args, 'outdir', builder_config.get('outdir', '.'))
    original_resolution = get_arg(args, 'original', builder_config.get('original'))
    modelname = get_arg(args, 'modelname', builder_config.get('modelname'))
    gridname = get_arg(args, 'gridname', builder_config.get('gridname'))
    rebuild = get_arg(args, 'rebuild', builder_config.get('rebuild', False))
    version = get_arg(args, 'version', builder_config.get('version'))
    verify = get_arg(args, 'verify', builder_config.get('verify', False))
    create_yaml = get_arg(args, 'yaml', builder_config.get('yaml', False))
    vert_coord = get_arg(args, 'vert_coord', builder_config.get('vert_coord'))

    # Ensure required arguments are present
    if model is None:
        raise ValueError("Model must be specified via --model or in the config file")
    if exp is None:
        raise ValueError("Experiment must be specified via --exp or in the config file")
    if source is None:
        raise ValueError("Source must be specified via --source or in the config file")

    # retrieve the data
    reader = Reader(catalog=catalog, model=model, exp=exp, source=source, loglevel=loglevel, areas=False, fix=fix)
    data = reader.retrieve()

    # Create GridBuilder instance
    grid_builder = GridBuilder(
        loglevel=loglevel,
        outdir=outdir, 
        original_resolution=original_resolution,
        model_name=modelname,
        grid_name=gridname,
        vert_coord=vert_coord
    )

    # Build the grid
    grid_builder.build(
        data, rebuild=rebuild,
        version=version, verify=verify,
        create_yaml=create_yaml)
    


"""Test for timmean method"""
import pytest
import numpy as np
from aqua.core.histogram import histogram

@pytest.fixture(scope='module')
def reader(ifs_tco79_long_fixFalse_reader):
    return ifs_tco79_long_fixFalse_reader

@pytest.fixture(scope='module')
def data(ifs_tco79_long_fixFalse_data):
    return ifs_tco79_long_fixFalse_data

@pytest.fixture(scope="module")
def data_2t(reader):
    """Retrieve only 2t variable (most commonly used)"""
    return reader.retrieve(var='2t')

@pytest.fixture(scope="module")
def data_ttr(reader):
    """Retrieve only ttr variable"""
    return reader.retrieve(var='ttr')

@pytest.mark.aqua
class TestTimmean():

    def test_timsum(self, reader, data):
        """Timmean test for sum operation"""
        summed = reader.timsum(data['2t'].isel(lon=0, lat=0), freq='3h')
        assert summed.shape == (1576,)
        assert summed[0] == data['2t'].isel(lon=0, lat=0, time=slice(0, 3)).sum()
        assert np.all(np.unique(summed.time.dt.hour) == np.arange(0, 24, 3))

        with pytest.raises(KeyError, match=r'hypertangent is not a statistic supported by AQUA'):
            reader.timstat(data['2t'], stat='hypertangent', freq='monthly', exclude_incomplete=True)

    @pytest.mark.parametrize('var', ['ttr'])
    def test_timmean_monthly(self, reader, data, var):
        """Timmean test for monthly aggregation"""
        avg = reader.timmean(data[var], freq='monthly')
        nmonths = len(np.unique(data.time.dt.month))
        unique, counts = np.unique(avg.time.dt.month, return_counts=True)
        assert avg.shape == (nmonths, 9, 18)
        assert len(unique) == nmonths
        assert all(counts == counts[0])

    @pytest.mark.parametrize('var', ['2t'])
    def test_timstd_allperiod(self, reader, data, var):
        """Timstd test for entire data period"""
        avg = reader.timstd(data[var])
        assert avg.shape == (9, 18)

    @pytest.mark.parametrize('var', ['2t', 'ttr'])
    def test_timstat_monthly_exclude_incomplete(self, reader, data, var):
        """Timmean test for monthly aggregation with excluded incomplete chunks"""
        avg = reader.timstat(data[var], stat='mean', freq='monthly', exclude_incomplete=True)
        unique, counts = np.unique(avg.time.dt.month, return_counts=True)
        assert avg.shape == (6, 9, 18)
        assert len(unique) == 6
        assert all(counts == counts[0])

    @pytest.mark.parametrize('var', ['2t', 'ttr'])
    def test_timmax_daily(self, reader, data, var):
        """Timmean test for daily aggregation"""
        avg = reader.timmax(data[var], freq='daily')
        unique, counts = np.unique(avg.time.dt.day, return_counts=True)
        assert avg.shape == (197, 9, 18)
        assert len(unique) == 31
        assert all(counts == np.array([7, 7, 7, 6, 6, 6, 6, 6, 6, 6,
                                       6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7,
                                       7, 7, 7, 7, 7, 7, 7, 6, 4]))
        
    def test_timstat_compare(self, reader, data_2t):
        """Time operations provide robust values"""
        minval = reader.timmin(data_2t['2t'], freq='daily')
        maxval = reader.timmax(data_2t['2t'], freq='daily')
        avg = reader.timmean(data_2t['2t'], freq='daily')
        
        assert (minval <= avg).all()
        assert (avg <= maxval).all()
    def test_timmin_yearly_exclude_incomplete(self, reader, data_ttr):
        """Timmean test for yearly aggregation with excluded incomplete chunks"""
        avg = reader.timmin(data_ttr, freq='yearly', exclude_incomplete=True)
        assert avg['ttr'].shape == (0, 9, 18)

    def test_timmean_yearly_center_time(self, reader, data_ttr):
        """Timmean test for yearly aggregation with center_time=True"""
        avg = reader.timmean(data_ttr, freq='yearly', center_time=True)
        assert avg['ttr'].shape == (1, 9, 18)
        assert avg['ttr'].time[0].values == np.datetime64('2020-07-02T00:00:00.000000000')

    def test_timmean_monthly_center_time(self, reader, data_2t):
        """Timmean test for monthly aggregation with center_time=True"""
        avg = reader.timmean(data_2t, freq='monthly', center_time=True)
        assert avg['2t'].shape == (8, 9, 18)
        assert avg['2t'].time[1].values == np.datetime64('2020-02-15T12:00:00.000000000')

    def test_timstd_daily_center_time(self, reader, data_2t):
        """Timmean test for daily aggregation with center_time=True and exclude_incomplete=True"""
        avg = reader.timstd(data_2t, freq='daily', center_time=True, exclude_incomplete=True)
        assert avg['2t'].shape == (197, 9, 18)
        assert avg['2t'].time[1].values == np.datetime64('2020-01-21T12:00:00.000000000')

    def test_timmean_pandas_accessor(self, reader, data_2t):
        """Timmean test for weekly aggregation based on pandas labels"""
        avg = data_2t.aqua.timmean(freq='W-MON')
        assert avg['2t'].shape == (29, 9, 18)

    def test_timmean_time_bounds(self, reader, data_2t):
        """Test for timmean method with time_bounds=True"""
        avg = reader.timmean(data_2t, freq='monthly', time_bounds=True)
        assert 'time_bnds' in avg
        assert avg['2t'].shape == (8, 9, 18)
        assert avg['time_bnds'].shape == (avg['2t'].shape[0], 2)
        assert np.all(avg['time_bnds'].isel(bnds=0) <= avg['time_bnds'].isel(bnds=1))

    def test_timmean_invalid_frequency(self, reader, data_2t):
        """Test for timmean method with an invalid frequency"""
        with pytest.raises(ValueError, match=r'Cant find a frequency to resample, using resample_freq=invalid not work, aborting!'):
            reader.timmean(data_2t, freq='invalid')

    def test_timstd_error(self, reader, data_2t):
        """Test for timstd method with a single time step"""
        single = data_2t.sel(time=data_2t.time[0])
        with pytest.raises(ValueError, match=r'Time dimension not found in the input data. Cannot compute timstd statistic'):
            avg = reader.timstat(single, stat='std', freq='monthly')

    def test_timstat_histogram(self, reader, data_2t):
        """Test histogram through timstat"""
        bins, range = 20, (250,330)

        #test passing a string
        hist1 = reader.timstat(data_2t['2t'], freq='monthly', stat='histogram', bins=bins, range=range, exclude_incomplete=True)
        # timhist passes a function
        hist2 = reader.timhist(data_2t['2t'], freq='monthly', bins=bins, range=range, exclude_incomplete=True)
        hist3 = reader.timstat(data_2t['2t'], freq='monthly', stat=histogram, bins=bins, range=range, exclude_incomplete=True)

        assert hist1['center_of_bin'].shape == hist2['center_of_bin'].shape
        assert hist1['center_of_bin'].shape == hist3['center_of_bin'].shape
        assert hist1.isel(time=2).sum().values == hist2.isel(time=2).sum().values
        assert hist2.isel(time=2).sum().values == hist3.isel(time=2).sum().values

        hist1 = reader.timhist(data_2t['2t'], bins=bins, range=range)
        hist2 = reader.histogram(data_2t['2t'], bins=bins, range=range)

        assert hist1.sum().values == hist2.sum().values

    def test_timmean_exclude_incomplete(self, reader, data_2t):
        """Timmean seasonal QS-FEB with exclude_incomplete=True"""
        da = data_2t['2t'].isel(lon=0, lat=0)

        # Use actual data range from 2020
        # Keep full months: Feb, Apr, May, Jun, Jul of 2020; omit March to make Q1 incomplete
        mask = (
            (da.time.dt.year == 2020) &
            (da.time.dt.month.isin([2, 4, 5, 6, 7]))
        )
        da_sel = da.sel(time=mask)
        avg = reader.timmean(da_sel, freq="QS-FEB", exclude_incomplete=True)

        # Only the complete quarter [May, Jun, Jul] should remain
        assert avg.time.dt.strftime("%Y-%m-%d").values[0] == "2020-05-01"

        # Expected value is the mean over May–Jul 2020 of the same series
        expected = da_sel.sel(time=da_sel.time.dt.month.isin([5, 6, 7])).mean().values
        assert float(avg.values) == float(expected)

    def test_timmean_exclude_incomplete_tcoords(self, reader, data_2t):
        """Test that exclude_incomplete mask coordinates align with resampled time axis"""
        da = data_2t['2t'].isel(lon=0, lat=0)

        # Get the averaged result 
        avg_with_mask = reader.timmean(da, freq='daily', exclude_incomplete=True)

        # Get what the resample time axis should be
        expected_time_axis = da.resample(time='1D').mean().time

        # The coordinates should match exactly (subset since incomplete are excluded)
        assert all(t in expected_time_axis.values for t in avg_with_mask.time.values), \
            "Masked result time coordinates should be a subset of the resampled time axis"

        # Verify alignment works without errors - this would fail with mismatched coords
        try:
            resampled = da.resample(time='1D').mean()
            # This operation requires matching coordinates
            aligned = resampled.sel(time=avg_with_mask.time)
            assert len(aligned) == len(avg_with_mask)
        except KeyError as e:
            pytest.fail(f"Coordinate alignment failed: {e}")


"""Module for aqua grid build"""
import os
from typing import Optional, Any
from smmregrid import GridInspector

from aqua.core.logger import log_configure, log_history
from .extragridbuilder import HealpixGridBuilder, RegularGridBuilder
from .extragridbuilder import UnstructuredGridBuilder, CurvilinearGridBuilder
from .extragridbuilder import GaussianRegularGridBuilder
from .gridentrymanager import GridEntryManager


class GridBuilder():
    """
    Class to build automatically grids from data sources.
    Currently supports HEALPix grids and can be extended for other grid types.
    """
    GRIDTYPE_REGISTRY = {
        'HEALPix': HealpixGridBuilder,
        'Regular': RegularGridBuilder,
        'Unstructured': UnstructuredGridBuilder,
        'Curvilinear': CurvilinearGridBuilder,
        'GaussianRegular': GaussianRegularGridBuilder,
        # Add more grid types here as needed
    }

    def __init__(
        self,
        outdir: str = '.',
        model_name: Optional[str] = None,
        grid_name: Optional[str] = None,
        original_resolution: Optional[str] = None,
        vert_coord: Optional[str] = None,
        loglevel: str = 'warning'
    ) -> None:
        """
        Initialize the GridBuilder with a reader instance.

        Args:
            outdir (str): The output directory for the grid files.
            model_name (str, optional): The name of the model, if different from the model argument.
            grid_name (str, optional): The name of the grid, to specify extra information in the grid file
            original_resolution (str, optional): The original resolution of the grid if using an interpolated source.
            vert_coord (str, optional): The vertical coordinate to consider for the grid build, to override the one detected by the GridInspector.
            loglevel (str, optional): The logging level for the logger. Defaults to 'warning'.
        """
        # store output directory
        self.outdir = outdir

        # store original resolution if necessary
        self.original_resolution = original_resolution

        # set model name
        self.model_name = model_name
        self.grid_name = grid_name

        # loglevel
        self.logger = log_configure(log_level=loglevel, log_name='GridBuilder')
        self.loglevel = loglevel

        # vertical coordinates to consider for the grid build for the 3d case.
        self.vert_coord = vert_coord

        # Initialize GridEntryManager to generate the grid file name and entry
        self.gem = GridEntryManager(
            model_name=self.model_name,
            grid_name=self.grid_name,
            original_resolution=self.original_resolution,
            loglevel=loglevel
        )

    def build(self, data, rebuild=False, version=None, verify=True, create_yaml=True):
        """
        Retrieve and build the grid data for all gridtypes available.

        Args:
            rebuild (bool): Whether to rebuild the grid file if it already exists. Defaults to False.
            fix (bool): Whether to fix the original source. Might be useful for some models. Defaults to False.
            version (int, optional): The version number to append to the grid file name. Defaults to None.
            verify (bool): Whether to verify the grid file after creation. Defaults to True.
            create_yaml (bool): Whether to create the grid entry in the grid file. Defaults to True.
        """
        gridtypes = GridInspector(data).get_gridtype()
        if not gridtypes:
            self.logger.error("No grid type detected, skipping grid build")
            self.logger.error("You can try to fix the source when calling the Reader() with the --fix flag")
            return
        self.logger.info("Build on %s gridtypes", len(gridtypes))
        for gridtype in gridtypes:
            self._build_gridtype(
                data, gridtype, rebuild=rebuild,
                version=version, verify=verify,
                create_yaml=create_yaml)

    def _build_gridtype(
        self,
        data: Any,
        gridtype: Any,
        rebuild: bool = False,
        version: Optional[int] = None,
        verify: bool = True,
        create_yaml: bool = True
    ) -> None:
        """
        Build the grid data based on the detected grid type.
        """
        self.logger.info("Detected grid type: %s", gridtype)
        kind = gridtype.kind
        self.logger.info("Grid type is: %s", kind)

        # access the class registry to get the builder class appropriate for the gridtype
        BuilderClass = self.GRIDTYPE_REGISTRY.get(kind)
        if not BuilderClass:
            raise NotImplementedError(f"Grid type {kind} is not implemented yet")
        self.logger.debug("Builder class: %s", BuilderClass)

        # vertical coordinate detection
        vert_coord = self.vert_coord if self.vert_coord else gridtype.vertical_dim
        self.logger.info("Detected vertical coordinate: %s", vert_coord)

        # Initialize the builder
        builder = BuilderClass(
            vert_coord=vert_coord, model_name=self.model_name, grid_name=self.grid_name,
            original_resolution=self.original_resolution, loglevel=self.loglevel
        )

        # data reduction. Load the data into memory for convenience.
        data3d = builder.data_reduction(data, gridtype, vert_coord).load()

        # add history attribute, get metadata from the attributes
        exp = data3d['mask'].attrs.get('AQUA_exp', None)
        source = data3d['mask'].attrs.get('AQUA_source', None)
        model = data3d['mask'].attrs.get('AQUA_model', None)
        log_history(data3d, msg=f'Gridfile generated with GridBuilder from {model}_{exp}_{source}')

        # store the data in a temporary netcdf file
        filename_tmp = f"{self.model_name}_{exp}_{source}.nc"
        self.logger.info("Saving tmp data in %s", filename_tmp)

        # configure attributes for the grid file
        data3d = builder.clean_attributes(data3d)
        data3d.to_netcdf(filename_tmp)

        # select the 2D slice of the data and detect the mask type
        # TODO: this will likely not work for 3d unstructured grids.
        # An alternative might be to check if the NaN are changing along the vertical coordinate.
        data2d = builder.select_2d_slice(data3d, vert_coord)
        masked = builder.detect_mask_type(data2d)
        self.logger.info("Masked type: %s", builder.masked)

        # get the basename and metadata for the grid file
        # need to read grid property from 2d data
        metadata = builder.get_metadata(data2d)
        aquagrid = metadata['aquagrid']
        cdogrid = metadata['cdogrid']
        self.logger.debug("Grid metadata: %s", metadata)

        # Initialize GridEntryManager for this gridtype
        basename = self.gem.get_basename(aquagrid, cdogrid, masked, vert_coord)

        if not cdogrid or masked:
            self.logger.warning("No CDO grid detected, or mask, need physical file")

            # create the base path for the grid file
            basepath = self.gem.get_versioned_basepath(self.outdir, basename, version=version)

            # check if the file already exists and clean it if needed
            filename = f"{basepath}.nc"
            if os.path.exists(filename):
                if rebuild:
                    self.logger.warning('File %s already exists, removing it', filename)
                    os.remove(filename)
                else:
                    self.logger.error("File %s already exists, skipping", filename)
                    return

            # write the grid file with the class specific method
            builder.write_gridfile(
                input_file=filename_tmp, output_file=filename, metadata=metadata
            )
        else:
            self.logger.warning("CDO grid %s detected without mask, skipping physical file creation for %s", cdogrid, basename)
            filename = cdogrid

        # cleanup
        self.logger.info('Removing temporary file %s', filename_tmp)
        os.remove(filename_tmp)

        # verify the creation of the weights
        if verify:
            builder.verify_weights(filename, metadata=metadata)

        # create the grid entry in the grid file
        if create_yaml:
            gridfile = self.gem.get_gridfilename(cdogrid, kind)
            self.logger.info("Creating grid entry in %s", gridfile)
            grid_entry_name = self.gem.create_grid_entry_name(aquagrid, cdogrid, masked, vert_coord)
            cdo_options = metadata.get('cdo_options') if metadata else None
            remap_method = metadata.get('remap_method') if metadata else None
            grid_block = self.gem.create_grid_entry_block(
                filename, horizontal_dims=gridtype.horizontal_dims, cdo_options=cdo_options, remap_method=remap_method,
                vert_coord=vert_coord
            )
            self.gem.create_grid_entry(gridfile, grid_entry_name, grid_block, rebuild=rebuild)



"""
Graphics module for Aqua.

This module contains the following files:
- single_map: For plotting a single map.
- multiple_maps: For plotting multiple maps.
- timeseres: For timeseries ans seasonal cycle plots.

The following functions are available:
- plot_single_map: Plot a single map.
- plot_single_map_diff: Plot a map as contour and
                        the difference with a reference as contourf
- plot_timeseries: Plot monthly and annual timeseries
- plot_seasonalcycle: Plot a seasonal cycle
- plot_maps: Plot multiple maps using plot_single_map
"""
from .gregory import plot_gregory_monthly, plot_gregory_annual
from .histogram import plot_histogram
from .hovmoller import plot_hovmoller
from .boxplot import boxplot
from .index_plot import index_plot, indexes_plot
from .single_map import plot_single_map, plot_single_map_diff
from .styles import ConfigStyle
from .timeseries import plot_timeseries, plot_seasonalcycle
from .multiple_maps import plot_maps, plot_maps_diff
from .vertical_profiles import plot_vertical_profile, plot_vertical_profile_diff
from .lat_lon_profiles import plot_lat_lon_profiles
from .seasonal_means import plot_seasonal_lat_lon_profiles
from .vertical_line import plot_vertical_lines


__all__ = ["plot_gregory_monthly", "plot_gregory_annual",
           "plot_histogram",
           "plot_hovmoller", "boxplot",
           "index_plot", "indexes_plot",
           "plot_single_map", "plot_single_map_diff",
           "ConfigStyle",
           "plot_timeseries", "plot_seasonalcycle",
           "plot_maps", "plot_maps_diff",
           "plot_vertical_profile", "plot_vertical_profile_diff",
           "plot_lat_lon_profiles",
           "plot_seasonal_lat_lon_profiles",
           "plot_vertical_lines"]


def activate_style(style: str = None,
                   loglevel: str = 'WARNING'):
    """Activate the style for graphical utilities
    Args:
        style (str): name of the style to load.
                     If not provided, it will be read from the configuration file.
        loglevel (str): logging level. Default is 'WARNING'.
    """
    ConfigStyle(style=style, loglevel=loglevel)
